# coding=utf-8
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################
import numpy as np
from onnx import onnx_pb as onnx_proto, helper
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE
from . import onnx_ops


class _OperatorNameContext:
    _history = []

    def __init__(self, oopb, basename):
        self.basename = basename
        self.oopb = oopb

    def __enter__(self):
        assert self.oopb.basename is None, "The previous context doesn't quit"
        self.oopb.basename = self.basename
        if len(_OperatorNameContext._history) > 0:
            self.oopb.upper_ctx = _OperatorNameContext._history[-1]
        _OperatorNameContext._history.append(self)
        return self.oopb

    def __exit__(self, type, value, traceback):
        assert self is _OperatorNameContext._history.pop()
        self.oopb.basename = None


class OnnxOperatorBuilder:
    def __init__(self, container, scope):
        self._container = container
        self._scope = scope
        self.basename = None
        # TODO: not all OnnxOperatorBuilder invocation is via as_default...
        # ... temporarily enable this for onnx_fx
        self.upper_ctx = None
        self.int32 = onnx_proto.TensorProto.INT32
        self.int64 = onnx_proto.TensorProto.INT64
        self.float = onnx_proto.TensorProto.FLOAT
        self.float16 = onnx_proto.TensorProto.FLOAT16
        self.double = onnx_proto.TensorProto.DOUBLE
        self.bool = onnx_proto.TensorProto.BOOL

    def as_default(self, basename):
        return _OperatorNameContext(self, basename)

    @property
    def upper_context(self):
        return self.upper_ctx

    def _process_inputs(self, inputs, name):
        if not isinstance(inputs, (list, tuple)):
            inputs = [inputs]
        ox_inputs = []
        for i_ in inputs:
            ox_n = i_
            if isinstance(i_, np.ndarray):
                ox_n = self._scope.get_unique_variable_name(name + '_i')
                self._container.add_initializer(
                    ox_n,
                    NP_TYPE_TO_TENSOR_TYPE[i_.dtype],
                    i_.shape,
                    i_.flatten()
                )
            elif isinstance(i_, (tuple, list)):
                ox_n = self._scope.get_unique_variable_name(name + i_[0])
                self._container.add_initializer(
                    ox_n,
                    i_[1],
                    i_[2].shape,
                    i_[2].flatten()
                )
            elif isinstance(ox_n, str):
                pass
            else:
                raise ValueError(
                    'Unknown type for ONNX initializer: {}'.format(type(ox_n)))
            ox_inputs.append(ox_n)

        return ox_inputs

    def _process_outputs(self, outputs, name):
        if outputs is None:
            ox_outputs = 1
        else:
            ox_outputs = outputs
        if isinstance(ox_outputs, int):
            ox_outputs = [self._scope.get_unique_variable_name(
                name + str(i_)) for i_ in range(ox_outputs)]
        elif isinstance(ox_outputs, (list, tuple)):
            pass
        else:
            raise ValueError(
                'Unknown type for outputs: {}'.format(type(ox_outputs)))
        return ox_outputs

    def _generate_name(self, type_or_func, name):
        base_name = (self.basename if self.basename else '') + '_'
        if name is not None:
            long_name = base_name + name
        else:
            if isinstance(type_or_func, str):
                suffix = type_or_func.lower()
            else:
                suffix = type_or_func.__name__[len('apply_'):]
            long_name = base_name + suffix
        return long_name

    @staticmethod
    def value_to_ndarray(value):
        if isinstance(value, (int, float, bool)):
            ty = np.int64
            if isinstance(value, float):
                ty = np.float32
            elif isinstance(value, bool):
                ty = np.bool_
            else:
                pass
            value = np.array(value).astype(ty)
        return value

    def _value_to_tensor(self, value, name, atleast_1d=False):
        value = type(self).value_to_ndarray(value)
        if isinstance(value, np.ndarray):
            if atleast_1d:
                value = np.atleast_1d(value)  # e.g. constant_of_shape() needs this
            lst = value.flatten().tolist()
            value = helper.make_tensor(name, NP_TYPE_TO_TENSOR_TYPE[value.dtype], value.shape, lst)
        return value

    def add_node(self, op_type, inputs, name=None, outputs=None, op_domain='', op_version=None, **attrs):
        if op_version is None:
            op_version = self._container.target_opset
        name = self._generate_name(op_type, name)
        ox_inputs = self._process_inputs(inputs, name)
        ox_outputs = self._process_outputs(outputs, name)
        self._container.add_node(op_type, ox_inputs, ox_outputs, op_domain, op_version,
                                 name=self._scope.get_unique_operator_name(name), **attrs)
        return ox_outputs[0] if outputs is None else ox_outputs

    def add_node_with_output(self, op_type, inputs, outputs, name, op_domain='', op_version=None, **attrs):
        if op_version is None:
            op_version = self._container.target_opset
        ox_inputs = self._process_inputs(inputs, name)
        self._container.add_node(
            op_type, ox_inputs, outputs, op_domain, op_version, name=name, **attrs)
        return outputs

    def apply_op(self, apply_func, inputs, name=None, outputs=None, **attrs):
        name = self._generate_name(apply_func, name)
        ox_inputs = self._process_inputs(inputs, name)
        ox_outputs = self._process_outputs(outputs, name)
        apply_func(self._scope, ox_inputs, ox_outputs, self._container,
                   operator_name=self._scope.get_unique_operator_name(name), **attrs)
        return ox_outputs[0] if outputs is None else ox_outputs

    def constant(self, name=None, value=None, outputs=None):
        c_name = self._scope.get_unique_variable_name(name or 'c')
        c_value = self._value_to_tensor(value, c_name)
        return self.apply_op(onnx_ops.apply_constant2, [], name, outputs, value=c_value)

    def constant_of_shape(self, inputs, name=None, outputs=None, value=None):
        if not isinstance(value, (int, float, bool)):
            raise ValueError("constant_of_shape requires 'value' to be a scalar")
        c_name = self._scope.get_unique_variable_name(name or 'cos')
        c_value = self._value_to_tensor(value, c_name, atleast_1d=True)
        return self.apply_op(onnx_ops.apply_constant_of_shape, inputs, name, outputs, value=c_value)

    def slice(self, inputs, name=None, outputs=None, starts=None, ends=None, axes=None, steps=None):
        return self.apply_op(onnx_ops.apply_slice2, inputs, name, outputs,
                             starts=starts, ends=ends, axes=axes, steps=steps)

    def loop(self, trip_count, cond, body, inputs, outputs, name=None):
        name = self._generate_name('loop', name)
        trip_count = '' if trip_count is None else trip_count
        cond_name = '' if cond is None else cond
        ox_inputs = self._process_inputs(inputs, name)
        ox_inputs = [trip_count, cond_name] + ox_inputs
        ox_outputs = outputs
        self._container.add_node(
            'Loop', ox_inputs, ox_outputs, op_version=1, name=name, body=body)
        return ox_outputs

    # !!!!CODE-AUTOGEN!!!! #
    # The following code was generated by ../update_ops.py

    def abs(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_abs, inputs, name, outputs)

    def add(self, inputs, name=None, outputs=None, axis=None, broadcast=None):
        return self.apply_op(onnx_ops.apply_add, inputs, name, outputs, axis=axis, broadcast=broadcast)

    def argmax(self, inputs, name=None, outputs=None, axis=0, keepdims=1, select_last_index=0):
        return self.apply_op(onnx_ops.apply_argmax, inputs, name, outputs, axis=axis, keepdims=keepdims,
                             select_last_index=select_last_index)

    def argmin(self, inputs, name=None, outputs=None, axis=0, keepdims=1, select_last_index=0):
        return self.apply_op(onnx_ops.apply_argmin, inputs, name, outputs, axis=axis, keepdims=keepdims,
                             select_last_index=select_last_index)

    def affine(self, inputs, name=None, outputs=None, alpha=1.0, beta=0.0):
        return self.apply_op(onnx_ops.apply_affine, inputs, name, outputs, alpha=alpha, beta=beta)

    def batch_norm(self, inputs, name=None, outputs=None, epsilon=None, is_test=None, momentum=None, spatial=None):
        return self.apply_op(onnx_ops.apply_batch_norm, inputs, name, outputs, epsilon=epsilon, is_test=is_test,
                             momentum=momentum, spatial=spatial)

    def cast(self, inputs, name=None, outputs=None, to=None):
        return self.apply_op(onnx_ops.apply_cast, inputs, name, outputs, to=to)

    def clip(self, inputs, name=None, outputs=None, max=None, min=None):
        return self.apply_op(onnx_ops.apply_clip, inputs, name, outputs, max=max, min=min)

    def concat(self, inputs, name=None, outputs=None, axis=0):
        return self.apply_op(onnx_ops.apply_concat, inputs, name, outputs, axis=axis)

    def conv(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_conv, inputs, name, outputs)

    def crop_height_width(self, inputs, name=None, outputs=None, top_border=0, bottom_border=0, left_border=0,
                          right_border=0):
        return self.apply_op(onnx_ops.apply_crop_height_width, inputs, name, outputs, top_border=top_border,
                             bottom_border=bottom_border, left_border=left_border, right_border=right_border)

    def div(self, inputs, name=None, outputs=None, axis=None, broadcast=None):
        return self.apply_op(onnx_ops.apply_div, inputs, name, outputs, axis=axis, broadcast=broadcast)

    def elu(self, inputs, name=None, outputs=None, alpha=1.0):
        return self.apply_op(onnx_ops.apply_elu, inputs, name, outputs, alpha=alpha)

    def equal(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_equal, inputs, name, outputs)

    def exp(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_exp, inputs, name, outputs)

    def floor(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_floor, inputs, name, outputs)

    def flatten(self, inputs, name=None, outputs=None, axis=1):
        return self.apply_op(onnx_ops.apply_flatten, inputs, name, outputs, axis=axis)

    def gather(self, inputs, name=None, outputs=None, axis=0):
        return self.apply_op(onnx_ops.apply_gather, inputs, name, outputs, axis=axis)

    def gemm(self, inputs, name=None, outputs=None, alpha=1.0, beta=1.0, transA=0, transB=0):
        return self.apply_op(onnx_ops.apply_gemm, inputs, name, outputs, alpha=alpha, beta=beta, transA=transA,
                             transB=transB)

    def greater(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_greater, inputs, name, outputs)

    def greater_or_equal(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_greater_or_equal, inputs, name, outputs)

    def less_or_equal(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_less_or_equal, inputs, name, outputs)

    def gru(self, inputs, name=None, outputs=None, output_seq=0, reset_after=0):
        return self.apply_op(onnx_ops.apply_gru, inputs, name, outputs, output_seq=output_seq, reset_after=reset_after)

    def hard_sigmoid(self, inputs, name=None, outputs=None, alpha=None, beta=None):
        return self.apply_op(onnx_ops.apply_hard_sigmoid, inputs, name, outputs, alpha=alpha, beta=beta)

    def identity(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_identity, inputs, name, outputs)

    def instance_norm(self, inputs, name=None, outputs=None, epsilon=1e-05):
        return self.apply_op(onnx_ops.apply_instance_norm, inputs, name, outputs, epsilon=epsilon)

    def inverse(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_inverse, inputs, name, outputs)

    def leaky_relu(self, inputs, name=None, outputs=None, alpha=0.01):
        return self.apply_op(onnx_ops.apply_leaky_relu, inputs, name, outputs, alpha=alpha)

    def less(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_less, inputs, name, outputs)

    def log(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_log, inputs, name, outputs)

    def lstm(self, inputs, name=None, outputs=None, output_seq=0):
        return self.apply_op(onnx_ops.apply_lstm, inputs, name, outputs, output_seq=output_seq)

    def matmul(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_matmul, inputs, name, outputs)

    def max(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_max, inputs, name, outputs)

    def mean(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_mean, inputs, name, outputs)

    def min(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_min, inputs, name, outputs)

    def mul(self, inputs, name=None, outputs=None, axis=None, broadcast=None):
        return self.apply_op(onnx_ops.apply_mul, inputs, name, outputs, axis=axis, broadcast=broadcast)

    def neg(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_neg, inputs, name, outputs)

    def normalization(self, inputs, name=None, outputs=None, axis=1, p=2):
        return self.apply_op(onnx_ops.apply_normalization, inputs, name, outputs, axis=axis, p=p)

    def not_op(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_not_op, inputs, name, outputs)

    def pad(self, inputs, name=None, outputs=None, mode=None, pads=None, value=None, onnx_type=1):
        return self.apply_op(onnx_ops.apply_pad, inputs, name, outputs, mode=mode, pads=pads, value=value,
                             onnx_type=onnx_type)

    def parametric_softplus(self, inputs, name=None, outputs=None, alpha=None, beta=None):
        return self.apply_op(onnx_ops.apply_parametric_softplus, inputs, name, outputs, alpha=alpha, beta=beta)

    def pow(self, inputs, name=None, outputs=None, axis=None, broadcast=None):
        return self.apply_op(onnx_ops.apply_pow, inputs, name, outputs, axis=axis, broadcast=broadcast)

    def prelu(self, inputs, name=None, outputs=None, slope=None):
        return self.apply_op(onnx_ops.apply_prelu, inputs, name, outputs, slope=slope)

    def range(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_range, inputs, name, outputs)

    def reciprocal(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_reciprocal, inputs, name, outputs)

    def reducesum(self, inputs, name=None, outputs=None, axes=None, keepdims=1, rank=0):
        return self.apply_op(onnx_ops.apply_reducesum, inputs, name, outputs, axes=axes, keepdims=keepdims, rank=rank)

    def relu(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_relu, inputs, name, outputs)

    def relu_6(self, inputs, name=None, outputs=None, zero_value=0.0):
        return self.apply_op(onnx_ops.apply_relu_6, inputs, name, outputs, zero_value=zero_value)

    def reshape(self, inputs, name=None, outputs=None, desired_shape=None):
        return self.apply_op(onnx_ops.apply_reshape, inputs, name, outputs, desired_shape=desired_shape)

    def resize(self, inputs, name=None, outputs=None, mode='nearest', coordinate_transformation_mode='asymmetric',
               scales=None):
        return self.apply_op(onnx_ops.apply_resize, inputs, name, outputs, mode=mode,
                             coordinate_transformation_mode=coordinate_transformation_mode, scales=scales)

    def rnn(self, inputs, name=None, outputs=None, output_seq=0):
        return self.apply_op(onnx_ops.apply_rnn, inputs, name, outputs, output_seq=output_seq)

    def shape(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_shape, inputs, name, outputs)

    def sigmoid(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_sigmoid, inputs, name, outputs)

    def softsign(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_softsign, inputs, name, outputs)

    def selu(self, inputs, name=None, outputs=None, alpha=1.673263, gamma=1.050701):
        return self.apply_op(onnx_ops.apply_selu, inputs, name, outputs, alpha=alpha, gamma=gamma)

    def softmax(self, inputs, name=None, outputs=None, axis=None):
        return self.apply_op(onnx_ops.apply_softmax, inputs, name, outputs, axis=axis)

    def scaled_tanh(self, inputs, name=None, outputs=None, alpha=None, beta=None):
        return self.apply_op(onnx_ops.apply_scaled_tanh, inputs, name, outputs, alpha=alpha, beta=beta)

    def slice2(self, inputs, name=None, outputs=None, starts=None, ends=None, axes=None, steps=None):
        return self.apply_op(onnx_ops.apply_slice2, inputs, name, outputs, starts=starts, ends=ends, axes=axes,
                             steps=steps)

    def split(self, inputs, name=None, outputs=None, split=None, axis=0):
        return self.apply_op(onnx_ops.apply_split, inputs, name, outputs, split=split, axis=axis)

    def sqrt(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_sqrt, inputs, name, outputs)

    def squeeze(self, inputs, name=None, outputs=None, axes=None, rank=0):
        return self.apply_op(onnx_ops.apply_squeeze, inputs, name, outputs, axes=axes, rank=rank)

    def sub(self, inputs, name=None, outputs=None, axis=None, broadcast=0):
        return self.apply_op(onnx_ops.apply_sub, inputs, name, outputs, axis=axis, broadcast=broadcast)

    def sum(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_sum, inputs, name, outputs)

    def tanh(self, inputs, name=None, outputs=None):
        return self.apply_op(onnx_ops.apply_tanh, inputs, name, outputs)

    def thresholded_relu(self, inputs, name=None, outputs=None, alpha=None):
        return self.apply_op(onnx_ops.apply_thresholded_relu, inputs, name, outputs, alpha=alpha)

    def tile(self, inputs, name=None, outputs=None, repeats=None):
        return self.apply_op(onnx_ops.apply_tile, inputs, name, outputs, repeats=repeats)

    def transpose(self, inputs, name=None, outputs=None, perm=None):
        return self.apply_op(onnx_ops.apply_transpose, inputs, name, outputs, perm=perm)

    def upsample(self, inputs, name=None, outputs=None, mode='nearest', coordinate_transformation_mode='asymmetric',
                 scales=None):
        return self.apply_op(onnx_ops.apply_upsample, inputs, name, outputs, mode=mode,
                             coordinate_transformation_mode=coordinate_transformation_mode, scales=scales)

    def unsqueeze(self, inputs, name=None, outputs=None, axes=None, rank=0):
        return self.apply_op(onnx_ops.apply_unsqueeze, inputs, name, outputs, axes=axes, rank=rank)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

from collections.abc import Mapping, MutableMapping
from collections import OrderedDict


class CaseInsensitiveDict(MutableMapping):
    def __init__(self, data=None, **kwargs):
        self._dict = OrderedDict()
        if data:
            self.update(data, **kwargs)

    def __setitem__(self, key, value):
        self._dict[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._dict[key.lower()][1]

    def __delitem__(self, key):
        del self._dict[key.lower()]

    def __iter__(self):
        return (key for key, _ in self._dict.values())

    def __len__(self):
        return len(self._dict)

    def lower_key_iteritems(self):
        """Like iteritems(), but with lowercase keys."""
        return (
            (lower_key, keyval[1])
            for lower_key, keyval
            in self._dict.items()
        )

    def __eq__(self, other):
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        return dict(self.lower_key_iteritems()) == dict(other.lower_key_iteritems())

    def copy(self):
        return CaseInsensitiveDict(self._dict.values())

    def __repr__(self):
        return str(dict(self.items()))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

# This file contains some high-level APIs for applying operations on variables specified by names. We should try our
# best to use those functions because they can produce ONNX operators according to the ONNX version specified in the
# `container` argument. Notice that those function behaviors are defined in a way very similar to ONNX-1.2.

import numpy as np
from onnx import onnx_pb as onnx_proto
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE


def _create_name_or_use_existing_one(scope, op_type, name):
    if name is None:
        return scope.get_unique_operator_name(op_type)
    else:
        return name


def _apply_unary_operation(scope, op_type, input_name, output_name, container, operator_name, **attrs):
    name = _create_name_or_use_existing_one(scope, op_type, operator_name)

    attrs['name'] = name
    if container.target_opset < 6:
        attrs['consumed_inputs'] = [0]
        op_version = 1
    else:
        op_version = 6

    container.add_node(op_type, input_name, output_name, op_version=op_version, **attrs)


def _apply_basic_numerical_operation(scope, op_type, input_names, output_name, container, operator_name,
                                     axis, broadcast):
    name = _create_name_or_use_existing_one(scope, op_type, operator_name)

    attrs = {}
    if container.target_opset < 7:
        # Before ONNX-1.2 (opset 7), broadcasting behavior is Caffe2-like.
        if axis is not None:
            attrs['axis'] = axis
        if broadcast is not None:
            attrs['broadcast'] = broadcast

        if container.target_opset < 6:
            attrs['consumed_inputs'] = [0, 0]
            op_version = 1
        else:
            op_version = 6
    else:
        # Since ONNX-1.2 (opset 7), broadcasting behavior is Numpy-like, so we don't need to specify any attributes
        op_version = 7

    container.add_node(op_type, input_names, output_name, op_version=op_version, name=name, **attrs)


def _apply_pointwise_operation(scope, op_type, input_names, output_name, container, operator_name):
    name = _create_name_or_use_existing_one(scope, op_type, operator_name)
    attrs = {}

    if container.target_opset < 6:
        attrs['consumed_inputs'] = [0] * len(input_names)
        op_version = 1
    elif container.target_opset < 8:
        op_version = 6
    else:
        if container.target_opset < 12 or op_type == 'Mean':
            op_version = 8
        else:
            op_version = 12

    container.add_node(op_type, input_names, output_name, op_version=op_version, name=name, **attrs)


def apply_abs(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Abs', input_name, output_name, container, operator_name=operator_name)


def apply_add(scope, input_names, output_name, container, operator_name=None, axis=None, broadcast=None):
    _apply_basic_numerical_operation(scope, 'Add', input_names, output_name, container, operator_name=operator_name,
                                     axis=axis, broadcast=broadcast)


def apply_argmax(scope, input_name, output_name, container, operator_name=None, axis=0, keepdims=1,
                 select_last_index=0):
    name = _create_name_or_use_existing_one(scope, 'ArgMax', operator_name)
    attrs = {'axis': axis, 'keepdims': keepdims}
    if container.target_opset < 11:
        op_version = 1
    elif container.target_opset < 12:
        op_version = 11
    else:
        op_version = 12
        attrs['select_last_index'] = select_last_index
    container.add_node('ArgMax', input_name, output_name, op_version=op_version, name=name, **attrs)


def apply_argmin(scope, input_name, output_name, container, operator_name=None, axis=0, keepdims=1,
                 select_last_index=0):
    name = _create_name_or_use_existing_one(scope, 'ArgMin', operator_name)
    attrs = {'axis': axis, 'keepdims': keepdims}
    if container.target_opset < 11:
        op_version = 1
    elif container.target_opset < 12:
        op_version = 11
    else:
        op_version = 12
        attrs['select_last_index'] = select_last_index
    container.add_node('ArgMin', input_name, output_name, op_version=op_version, name=name, **attrs)


def apply_affine(scope, input_name, output_name, container, operator_name=None, alpha=1., beta=0.):
    if container.target_opset < 9:
        op_type = 'Affine'
        name = _create_name_or_use_existing_one(scope, 'Affine', operator_name)
        attrs = {'name': name, 'alpha': alpha, 'beta': beta}
        container.add_node(op_type, input_name, output_name, **attrs)
    else:
        name = _create_name_or_use_existing_one(scope, 'Affine', operator_name)
        # Define a and b.
        aName = scope.get_unique_variable_name(name + '_alpha')
        container.add_initializer(aName, onnx_proto.TensorProto.FLOAT, [1], [alpha])
        bName = scope.get_unique_variable_name(name + '_beta')
        container.add_initializer(bName, onnx_proto.TensorProto.FLOAT, [1], [beta])

        # Compute Z = a * X, where X is the original input.
        zName = scope.get_unique_variable_name(name + '_scaled')
        apply_mul(scope, [aName, input_name], zName, container)

        # Compute Y = Z + b, where Y is the final output.
        apply_add(scope, [zName, bName], output_name, container)


def apply_batch_norm(scope, input_names, output_names, container, operator_name=None,
                     epsilon=None, is_test=None, momentum=None, spatial=None):
    name = _create_name_or_use_existing_one(scope, 'BatchNormalization', operator_name)
    attrs = {'name': name, 'epsilon': epsilon, 'momentum': momentum}

    if container.target_opset < 9:
        attrs['spatial'] = spatial
    if container.target_opset < 7:
        attrs['is_test'] = is_test

    if container.target_opset < 6:
        attrs['consumed_inputs'] = [0] * len(input_names)
        if len(input_names) > 3:
            attrs['consumed_inputs'][3] = 1
        if len(input_names) > 4:
            attrs['consumed_inputs'][4] = 2
        op_version = 1
    elif container.target_opset < 7:
        op_version = 6
    elif container.target_opset < 9:
        op_version = 7
    else:
        op_version = 9

    container.add_node('BatchNormalization', input_names, output_names, op_version=op_version, **attrs)


def apply_cast(scope, input_name, output_name, container, operator_name=None, to=None):
    '''
    :param to: enum defined in ONNX TensorProto.DataType, for example, TensorProto.FLOAT and TensorProto.INT64.
    '''
    name = _create_name_or_use_existing_one(scope, 'Cast', operator_name)
    attrs = {'name': name}

    d = onnx_proto.TensorProto.DataType.DESCRIPTOR
    allowed_type_name_and_type_enum_pairs = {v.number: k for k, v in d.values_by_name.items()}
    if to not in allowed_type_name_and_type_enum_pairs:
        raise ValueError('Attribute "to" must be one of %s' % allowed_type_name_and_type_enum_pairs.keys())

    if container.target_opset < 9:
        if to in [onnx_proto.TensorProto.STRING, onnx_proto.TensorProto.COMPLEX64, onnx_proto.TensorProto.COMPLEX128]:
            raise ValueError('Attribute "to" cannot correspond to a String or Complex TensorProto type.')

        if container.target_opset < 6:
            # Convert enum to string, for example, TensorProto.INT64 to 'INT64'
            attrs['to'] = allowed_type_name_and_type_enum_pairs[to]
            op_version = 1
        else:
            # Enum, for example, TensorProto.INT64
            attrs['to'] = to
            op_version = 6
    else:
        # Enum value, for example, TensorProto.INT64
        # String casting is supported in opset 9
        if to in [onnx_proto.TensorProto.COMPLEX64, onnx_proto.TensorProto.COMPLEX128]:
            raise ValueError('Attribute "to" cannot correspond to a Complex TensorProto type.')
        attrs['to'] = to
        op_version = 9

    container.add_node('Cast', input_name, output_name, op_version=op_version, **attrs)


def apply_clip(scope, input_name, output_name, container, operator_name=None, max=None, min=None):
    name = _create_name_or_use_existing_one(scope, 'Clip', operator_name)
    attrs = {'name': name}

    if container.target_opset < 11:
        if max is not None:
            attrs['max'] = float(max)
        if min is not None:
            attrs['min'] = float(min)

        if container.target_opset < 6:
            attrs['consumed_inputs'] = [0]
            op_version = 1
        else:
            op_version = 6

        container.add_node('Clip', input_name, output_name, op_version=op_version, **attrs)
    else:
        if container.target_opset < 12:
            op_version = 11
        else:
            op_version = 12
        if min is None and max is not None:
            raise RuntimeError("Operator 'Clip': min must be specified if max is.")
        inputs = [input_name]

        if min is not None:
            if isinstance(min, (np.ndarray, float, int)):
                # add initializer
                if isinstance(min, np.ndarray):
                    if len(min.shape) == 0:
                        min = [min]
                    elif min.shape == (1,):
                        min = list(min[0]) if hasattr(min[0], '__iter__') else list(min)
                    else:
                        raise RuntimeError("min must be an array of one element.")
                else:
                    min = [min]

                # container in sklearn-onnx stores the computation type in
                # container.dtype.
                min_name = scope.get_unique_variable_name('clip_min')
                if op_version < 12:
                    min = np.array(min, dtype=getattr(container, 'dtype', np.float32))
                    container.add_initializer(min_name, getattr(container, 'proto_dtype',
                                                                onnx_proto.TensorProto.FLOAT), [], [min[0]])
                else:
                    min = np.array(min)
                    container.add_initializer(min_name, NP_TYPE_TO_TENSOR_TYPE[min.dtype], [], [min[0]])
                min = min_name
            if isinstance(min, str):
                inputs.append(min)
            else:
                raise RuntimeError("Parameter 'min' must be a string or a float.")

        if max is not None:
            if min is None:
                raise RuntimeError("Parameter 'min' must be specified if 'max' is.")
            if isinstance(max, (np.ndarray, float, int)):
                # add initializer
                if isinstance(max, np.ndarray):
                    if len(max.shape) == 0:
                        max = [max]
                    elif max.shape == (1,):
                        max = list(max[0]) if hasattr(max[0], '__iter__') else list(max)
                    else:
                        raise RuntimeError("max must be an array of one element.")
                else:
                    max = [max]

                max_name = scope.get_unique_variable_name('clip_max')
                if op_version < 12:
                    max = np.array(max, dtype=getattr(container, 'dtype', np.float32))
                    container.add_initializer(max_name, getattr(container, 'proto_dtype',
                                                                onnx_proto.TensorProto.FLOAT), [], [max[0]])
                else:
                    max = np.array(max)
                    container.add_initializer(max_name, NP_TYPE_TO_TENSOR_TYPE[max.dtype], [], [max[0]])
                max = max_name
            if isinstance(max, str):
                inputs.append(max)
            else:
                raise RuntimeError("Parameter 'max' must be a string or a float.")

        container.add_node('Clip', inputs, output_name, op_version=op_version,
                           **attrs)


def apply_concat(scope, input_names, output_name, container, operator_name=None, axis=0):
    name = _create_name_or_use_existing_one(scope, 'Concat', operator_name)

    if container.target_opset < 4:
        op_version = 1
    elif container.target_opset < 11:
        op_version = 4
    else:
        op_version = 11

    container.add_node('Concat', input_names, output_name, op_version=op_version, name=name, axis=axis)


def apply_constant(scope, output_name, container, operator_name=None, value=None):
    name = _create_name_or_use_existing_one(scope, 'Constant', operator_name)

    if value is None:
        raise ValueError('Attribute "value" is a required argument.')

    if container.target_opset < 9:
        op_version = 1
    elif container.target_opset < 11:
        op_version = 9
    elif container.target_opset < 12:
        op_version = 11
    else:
        op_version = 12

    if op_version < 12:
        attrs = {'name': name, 'value': value}
    else:
        if isinstance(value, float):
            attrs = {'name': name, 'value_float': value}
        elif isinstance(value, int):
            attrs = {'name': name, 'value_int': value}
        elif isinstance(value, str):
            attrs = {'name': name, 'value_string': value}
        else:
            attrs = {'name': name, 'value': value}

    container.add_node('Constant', [], output_name, op_version=op_version, **attrs)


def apply_constant2(scope, input_names, output_name, container, operator_name=None, value=None):
    assert len(input_names) == 0  # only a placeholder to standardize the argument list.
    return apply_constant(scope, output_name, container, operator_name, value)


def apply_constant_of_shape(scope, input_names, output_name, container, operator_name=None, value=None):
    name = _create_name_or_use_existing_one(scope, 'ConstantOfShape', operator_name)
    container.add_node('ConstantOfShape', input_names, output_name, name=name, op_version=9, value=value)


def apply_conv(scope, input_names, output_name, container, operator_name=None, **attrs):
    name = _create_name_or_use_existing_one(scope, 'Conv', operator_name)

    if container.target_opset < 11:
        op_version = 1
    else:
        op_version = 11

    container.add_node('Conv', input_names, output_name, name=name, op_version=op_version, **attrs)


def apply_crop_height_width(scope, input_name, output_name, container, operator_name=None,
                            top_border=0, bottom_border=0, left_border=0, right_border=0):
    name = scope.get_unique_operator_name('CropHeightWidth')
    if container.target_opset < 9:
        # If operator set < 9, we can use the experimental Crop in ONNX.
        attrs = {'name': name, 'border': [left_border, top_border, right_border, bottom_border]}
        container.add_node('Crop', input_name, output_name, **attrs)
    else:
        # The experimental Crop in ONNX is removed after operator set 9, so we
        # switch to ONNX DynamicSlice operator.

        # CoreML only crops H- and W-axes.
        axes = [2, 3]
        axes_name = scope.get_unique_variable_name(name + '_axes')
        container.add_initializer(axes_name, onnx_proto.TensorProto.INT64,
                                  [len(axes)], axes)

        # Number of cropped pixels is the starting index of the remained region.
        starts = [top_border, left_border]
        starts_name = scope.get_unique_variable_name(name + '_starts')
        container.add_initializer(starts_name, onnx_proto.TensorProto.INT64,
                                  [len(starts)], starts)

        # First we assume no cropping is needed at the end of those axes.
        # We will change this right below depending on Crop's configuration.
        ends = [np.iinfo(np.int64).max] * 2

        # Crop n pixel means the end index (exclusive) is -n. Note that indexing
        # system is zero-based.
        if bottom_border > 0:
            ends[0] = -bottom_border
        if right_border > 0:
            ends[1] = -right_border

        # Add the adjusted ends.
        ends_name = scope.get_unique_variable_name(name + '_ends')
        container.add_initializer(ends_name, onnx_proto.TensorProto.INT64,
                                  [len(ends)], ends)

        # Collect all input names as a list because DynamicSlice has multiple inputs.
        input_list = [input_name, starts_name, ends_name, axes_name]
        container.add_node('DynamicSlice', input_list, output_name, op_version=9)


def apply_div(scope, input_names, output_name, container, operator_name=None, axis=None, broadcast=None):
    _apply_basic_numerical_operation(scope, 'Div', input_names, output_name, container, operator_name=operator_name,
                                     axis=axis, broadcast=broadcast)


def apply_elu(scope, input_name, output_name, container, operator_name=None, alpha=1.0):
    _apply_unary_operation(scope, 'Elu', input_name, output_name, container, operator_name, alpha=alpha)


def apply_equal(scope, input_names, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'equal', operator_name)
    if container.target_opset < 7:
        op_version = 1
    elif container.target_opset < 9:
        op_version = 7
    else:
        op_version = 9
    container.add_node('Equal', input_names, output_name, name=name, op_version=op_version)


def apply_exp(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Exp', input_name, output_name, container, operator_name=operator_name)


def apply_floor(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Floor', input_name, output_name, container, operator_name=operator_name)


def apply_flatten(scope, input_name, output_name, container, operator_name=None, axis=1):
    name = _create_name_or_use_existing_one(scope, 'Flatten', operator_name)
    if container.target_opset < 9:
        op_version = 1
    elif container.target_opset < 11:
        op_version = 9
    else:
        op_version = 11
    container.add_node('Flatten', input_name, output_name, name=name, op_version=op_version, axis=axis)


def apply_gather(scope, input_names, output_name, container, operator_name=None, axis=0):
    name = _create_name_or_use_existing_one(scope, 'Gather', operator_name)
    if container.target_opset < 11:
        op_version = 1
    else:
        op_version = 11

    container.add_node('Gather', input_names, output_name, name=name, op_version=op_version, axis=axis)


def apply_gemm(scope, input_name, output_name, container, operator_name=None, alpha=1.0, beta=1.0,
               transA=0, transB=0):
    """
    Applies operator `gemm <https://github.com/onnx/onnx/blob/master/docs/Operators.md#gemm>`.
    """
    name = _create_name_or_use_existing_one(scope, 'Gemm', operator_name)
    attrs = {'alpha': alpha, 'beta': beta, 'transA': transA, 'transB': transB}
    if container.target_opset < 5:
        attrs['op_version'] = 1
        attrs['broadcast'] = 1
    elif container.target_opset < 7:
        attrs['op_version'] = 6
        attrs['broadcast'] = 1
    elif container.target_opset < 11:
        attrs['op_version'] = 7
    else:
        attrs['op_version'] = 11

    container.add_node('Gemm', input_name, output_name, name=name, **attrs)


def apply_greater(scope, input_names, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Greater', operator_name)
    if container.target_opset < 7:
        op_version = 1
    elif container.target_opset < 9:
        op_version = 7
    else:
        op_version = 9

    container.add_node('Greater', input_names, output_name, name=name, op_version=op_version)


def _convert_compare_equal(scope, input_names, output_name, container, operator_name, tf_op_string, onnx_op_string_rev,
                           onnx_op_string):
    if container.target_opset < 7:
        raise ValueError(tf_op_string + " op is not supported for opset < 7")
    elif container.target_opset < 9:
        op_version = 7
    elif container.target_opset < 12:
        op_version = 9
    else:
        op_version = 12
    name = _create_name_or_use_existing_one(scope, tf_op_string, operator_name)
    if op_version < 9:
        compare_input_0 = scope.get_unique_variable_name(name + '_input_0_cast')
        container.add_node('Cast', [input_names[0]], compare_input_0, name=name + '_input_0_cast', to=1)
        compare_input_1 = scope.get_unique_variable_name(name + '_input_1_cast')
        container.add_node('Cast', [input_names[1]], compare_input_1, name=name + '_input_1_cast', to=1)
        less_out = scope.get_unique_variable_name(name + '_less_out')
        container.add_node(onnx_op_string_rev, [compare_input_0, compare_input_1], less_out,
                           name=name + '_' + onnx_op_string_rev.lower(),
                           op_version=op_version)
        container.add_node('Not', less_out, output_name, name=name + '_not')
    elif op_version < 12:
        compare_node = scope.get_unique_variable_name(name + '_compare_node')
        container.add_node(onnx_op_string_rev, input_names, compare_node,
                           name=name + '_' + onnx_op_string_rev.lower(),
                           op_version=op_version)
        container.add_node('Not', [compare_node], output_name, name=name)
    else:
        container.add_node(onnx_op_string, input_names, output_name,
                           name=name + '_' + onnx_op_string_rev.lower(), op_version=op_version)


def apply_greater_or_equal(scope, input_names, output_name, container, operator_name=None):
    _convert_compare_equal(scope, input_names, output_name, container, operator_name, 'GreaterEqual', 'Less',
                           'GreaterOrEqual')


def apply_less_or_equal(scope, input_names, output_name, container, operator_name=None):
    _convert_compare_equal(scope, input_names, output_name, container, operator_name, 'LessEqual', 'Greater',
                           'LessOrEqual')


def apply_gru(scope, input_names, output_names, container, operator_name=None, output_seq=0, reset_after=0, **attrs):
    name = _create_name_or_use_existing_one(scope, 'GRU', operator_name)
    if container.target_opset < 3:
        op_version = 1
        attrs['output_sequence'] = 1 if output_seq else 0
    else:
        attrs['linear_before_reset'] = 1 if reset_after else 0
        if container.target_opset <= 5:
            attrs['output_sequence'] = 1 if output_seq else 0
            op_version = 3
        elif container.target_opset <= 13:
            op_version = 7
        else:
            op_version = 14

    container.add_node('GRU', input_names, output_names, name=name, op_version=op_version, **attrs)


def apply_hard_sigmoid(scope, input_name, output_name, container, operator_name=None, alpha=None, beta=None):
    _apply_unary_operation(scope, 'HardSigmoid', input_name, output_name, container, operator_name,
                           alpha=alpha, beta=beta)


def apply_identity(scope, input_name, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Identity', operator_name)
    container.add_node('Identity', input_name, output_name, name=name)


def apply_instance_norm(scope, input_names, output_name, container, operator_name=None, epsilon=1e-5):
    name = _create_name_or_use_existing_one(scope, 'InstanceNormalization', operator_name)
    attrs = {'name': name, 'epsilon': epsilon}

    if container.target_opset < 2:
        attrs['consumed_inputs'] = [0] * len(input_names)
        op_version = 1
    else:
        op_version = 6

    container.add_node('InstanceNormalization', input_names, output_name, op_version=op_version, **attrs)


def apply_inverse(scope, input_name, output_name, container, operator_name=None):
    if container.target_opset < 12:
        raise ValueError("tf op MatrixInverse is not supported for opset < 12")
    else:
        op_version = 12
    name = _create_name_or_use_existing_one(scope, 'Inverse', operator_name)
    container.add_node('Inverse', input_name, output_name, name=name, op_version=op_version)


def apply_leaky_relu(scope, input_name, output_name, container, operator_name=None, alpha=0.01):
    _apply_unary_operation(scope, 'LeakyRelu', input_name, output_name, container, operator_name, alpha=alpha)


def apply_less(scope, input_names, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Less', operator_name)
    if container.target_opset < 7:
        op_version = 1
    elif container.target_opset < 9:
        op_version = 7
    else:
        op_version = 9

    container.add_node('Less', input_names, output_name, name=name, op_version=op_version)


def apply_log(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Log', input_name, output_name, container, operator_name=operator_name)


def apply_lstm(scope, input_names, output_names, container, operator_name=None, output_seq=0, **attrs):
    name = _create_name_or_use_existing_one(scope, 'LSTM', operator_name)
    if container.target_opset <= 6:
        attrs['output_sequence'] = 1 if output_seq else 0
        op_version = 1
    elif container.target_opset <= 13:
        op_version = 7
    else:
        op_version = 14
    container.add_node('LSTM', input_names, output_names, name=name, op_version=op_version, **attrs)


def apply_matmul(scope, input_names, output_name, container, operator_name=None):
    op_type = 'MatMul'
    name = _create_name_or_use_existing_one(scope, op_type, operator_name)
    if container.target_opset <= 9:
        op_version = 1
    else:
        op_version = 9
    container.add_node(op_type, input_names, output_name, op_version=op_version, name=name)


def apply_max(scope, input_names, output_name, container, operator_name=None):
    _apply_pointwise_operation(scope, 'Max', input_names, output_name, container, operator_name)


def apply_mean(scope, input_names, output_name, container, operator_name=None):
    _apply_pointwise_operation(scope, 'Mean', input_names, output_name, container, operator_name)


def apply_min(scope, input_names, output_name, container, operator_name=None):
    _apply_pointwise_operation(scope, 'Min', input_names, output_name, container, operator_name)


def apply_mul(scope, input_names, output_name, container, operator_name=None, axis=None, broadcast=None):
    _apply_basic_numerical_operation(scope, 'Mul', input_names, output_name, container, operator_name=operator_name,
                                     axis=axis, broadcast=broadcast)


def apply_neg(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Neg', input_name, output_name, container, operator_name)


def apply_normalization(scope, input_name, output_name, container, operator_name=None, axis=1, p=2):
    name = _create_name_or_use_existing_one(scope, 'LpNormalization', operator_name)
    container.add_node('LpNormalization', input_name, output_name, name=name, p=p, axis=axis)


def apply_not_op(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Not', input_name, output_name, container, operator_name)


def apply_pad(scope, input_name, output_name, container, operator_name=None, mode=None, pads=None, value=None,
              onnx_type=onnx_proto.TensorProto.FLOAT):
    name = _create_name_or_use_existing_one(scope, 'Pad', operator_name)
    attrs = {'name': name}
    inputs = input_name if isinstance(input_name, list) else [input_name]

    if mode is not None:
        attrs['mode'] = mode

    if container.target_opset < 11:
        if isinstance(pads, str):
            raise ValueError("Dynamic pad is not supported for opset < 11.")
        if value is not None:
            attrs['value'] = value
        if container.target_opset < 2:
            attrs['paddings'] = pads
            op_version = 1
        else:
            attrs['pads'] = pads
            op_version = 2
    else:
        op_version = 11
        if isinstance(pads, str):
            inputs.append(pads)
        else:
            pads_name = scope.get_unique_variable_name(name + '_pads')
            container.add_initializer(pads_name, onnx_proto.TensorProto.INT64, [len(pads)], pads)
            inputs.append(pads_name)
        if value is not None:
            value_name = scope.get_unique_variable_name(name + '_value')
            container.add_initializer(value_name, onnx_type, [], [value])
            inputs.append(value_name)

    container.add_node('Pad', inputs, output_name, op_version=op_version, **attrs)


def apply_parametric_softplus(scope, input_name, output_name, container, operator_name=None, alpha=None, beta=None):
    if alpha is None:
        alpha = [1.0]
    if beta is None:
        beta = [0.]

    name = _create_name_or_use_existing_one(scope, 'ParametricSoftplus', operator_name)
    if container.target_opset < 9:
        if len(alpha) != 1 or len(beta) != 1:
            raise ValueError('alpha and beta must be 1-element lists')
        op_type = 'ParametricSoftplus'
        attrs = {'name': name, 'alpha': alpha[0], 'beta': beta[0]}
        container.add_node(op_type, input_name, output_name, **attrs)
    else:
        # Define three scalars: a, b, 1.
        aName = scope.get_unique_variable_name(name + '_alpha')
        aShape = [len(alpha)] if len(alpha) == 1 else [len(alpha), 1, 1]
        container.add_initializer(aName, onnx_proto.TensorProto.FLOAT, aShape, alpha)
        bShape = [len(beta)] if len(beta) == 1 else [len(beta), 1, 1]
        bName = scope.get_unique_variable_name(name + '_beta')
        container.add_initializer(bName, onnx_proto.TensorProto.FLOAT, bShape, beta)
        oneName = scope.get_unique_variable_name(name + '_one')
        container.add_initializer(oneName, onnx_proto.TensorProto.FLOAT, [1], [1.])

        # c = b * x
        cName = scope.get_unique_variable_name(name + '_c')
        apply_mul(scope, [input_name, bName], cName, container)

        # d = exp(c)
        dName = scope.get_unique_variable_name(name + '_d')
        apply_exp(scope, cName, dName, container)

        # e = 1 + d
        eName = scope.get_unique_variable_name(name + '_e')
        apply_add(scope, [dName, oneName], eName, container)

        # f = log(e)
        fName = scope.get_unique_variable_name(name + '_f')
        apply_log(scope, eName, fName, container)

        # g = a * f
        apply_mul(scope, [fName, aName], output_name, container)


def apply_pow(scope, input_names, output_name, container, operator_name=None, axis=None, broadcast=None):
    name = _create_name_or_use_existing_one(scope, 'Pow', operator_name)

    attrs = {'name': name}
    if container.target_opset < 7:
        # Before ONNX-1.2, broadcasting behavior is Caffe2-like.
        if axis is not None:
            attrs['axis'] = axis
        if broadcast is not None:
            attrs['broadcast'] = broadcast
        op_version = 1
    elif container.target_opset < 12:
        # Since ONNX-1.2, broadcasting behavior is Numpy-like, so we don't need to specify any attributes
        op_version = 7
    else:
        op_version = 12

    container.add_node('Pow', input_names, output_name, op_version=op_version, **attrs)


def apply_prelu(scope, input_name, output_name, container, operator_name=None, slope=None):
    name = _create_name_or_use_existing_one(scope, 'PRelu', operator_name)
    slope_tensor_name = scope.get_unique_variable_name('slope')
    s_shape = slope.shape
    if container.target_opset < 7:
        s_shape = [len(slope.flatten())]
    container.add_initializer(slope_tensor_name, onnx_proto.TensorProto.FLOAT, s_shape, slope.flatten())

    if container.target_opset < 6:
        container.add_node('PRelu', [input_name, slope_tensor_name], output_name, op_version=1, name=name,
                           consumed_inputs=[0, 0])
    else:
        if container.target_opset < 7:
            op_version = 6
        elif container.target_opset < 9:
            op_version = 7
        else:
            # opset 9 supports unidirectional broadcasting
            op_version = 9

        container.add_node('PRelu', [input_name, slope_tensor_name], output_name, op_version=op_version, name=name)


def apply_range(scope, input_name, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Range', operator_name)
    container.add_node('Range', input_name, output_name, op_version=11, name=name)


def apply_reciprocal(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Reciprocal', input_name, output_name, container, operator_name=operator_name)


# Some old ORT supports axis < 0 case, so put rank=0 as default.
def apply_reducesum(scope, input_name, output_name, container, operator_name=None, axes=None, keepdims=1, rank=0):
    name = _create_name_or_use_existing_one(scope, 'ReduceSum', operator_name)
    if axes is None:
        axes = []
    if container.target_opset < 13:
        if container.target_opset < 11:
            op_version = 1
            axes = [axis if axis >= 0 else axis + rank for axis in axes]
        else:
            op_version = 11
        container.add_node('ReduceSum', input_name, output_name, name=name,
                           op_version=op_version, axes=axes, keepdims=keepdims)
    else:
        if not isinstance(input_name, list):
            input_name = [input_name]
        op_version = 13
        if isinstance(axes, str):
            container.add_node('ReduceSum', input_name + [axes], output_name,
                               op_version=op_version, name=name, keepdims=keepdims)
        elif axes is None or len(axes) == 0:
            container.add_node('ReduceSum', input_name, output_name,
                               op_version=op_version, name=name, keepdims=keepdims)
        else:
            axes_name = scope.get_unique_variable_name(name + '_reducesum')
            container.add_initializer(axes_name, onnx_proto.TensorProto.INT64, [len(axes)], axes)
            container.add_node('ReduceSum', input_name + [axes_name], output_name,
                               op_version=op_version, name=name, keepdims=keepdims)


def apply_relu(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Relu', input_name, output_name, container, operator_name)


def apply_relu_6(scope, input_name, output_name, container, operator_name=None, zero_value=0.0):
    name_relu = _create_name_or_use_existing_one(scope, 'relu', operator_name)
    name_relu_op = _create_name_or_use_existing_one(scope, 'relu6', operator_name)
    apply_relu(scope, input_name, name_relu, container, name_relu_op+'_relu')
    apply_clip(scope, name_relu, output_name, container, name_relu_op + '_clip', zero_value+6, zero_value)


def apply_reshape(scope, input_name, output_name, container, operator_name=None, desired_shape=None):
    if not isinstance(desired_shape, str) and len(list(i for i in desired_shape if i is not None and i < 0)) > 1:
        raise ValueError('There can only be one -1 in the targeted shape of a Reshape but got %s' % desired_shape)

    name = _create_name_or_use_existing_one(scope, 'Reshape', operator_name)

    if container.target_opset < 5:
        container.add_node('Reshape', input_name, output_name, op_version=1, name=name, shape=desired_shape,
                           consumed_inputs=[0])
    else:
        if isinstance(desired_shape, str):
            desired_shape_name = desired_shape
        else:
            desired_shape_name = scope.get_unique_variable_name('shape_tensor')
            container.add_initializer(desired_shape_name, onnx_proto.TensorProto.INT64, [len(desired_shape)],
                                      desired_shape)

        # Create ONNX Reshape operator
        if isinstance(input_name, list):
            input_name.append(desired_shape_name)
        else:
            input_name = [input_name, desired_shape_name]
        container.add_node('Reshape', input_name, output_name, op_version=5, name=name)


def apply_resize(scope, input_name, output_name, container, operator_name=None, mode='nearest',
                 coordinate_transformation_mode='asymmetric', scales=None):
    '''
    :param mode: "nearest" or "linear"
    :param scales: a float tensor for scaling (upsampling or downsampling) all input dimensions
    '''
    name = _create_name_or_use_existing_one(scope, 'Resize', operator_name)
    attrs = {'name': name}
    attrs['mode'] = mode.lower()

    inputs = [input_name]

    if container.target_opset < 11:
        op_version = 10
    else:
        op_version = 11
        roi_tensor_name = scope.get_unique_variable_name(name + '_roi')
        roi = [0.0] * len(scales) + [1.0] * len(scales)
        container.add_initializer(roi_tensor_name, onnx_proto.TensorProto.FLOAT, [2 * len(scales)], roi)
        inputs.append(roi_tensor_name)
        attrs['coordinate_transformation_mode'] = coordinate_transformation_mode
        if attrs['mode'] == 'nearest':
            attrs['nearest_mode'] = 'floor'

    scales_tensor_name = scope.get_unique_variable_name(name + '_scales')
    container.add_initializer(scales_tensor_name, onnx_proto.TensorProto.FLOAT, [len(scales)], scales)
    inputs.append(scales_tensor_name)
    container.add_node('Resize', inputs, output_name, op_version=op_version, **attrs)


def apply_rnn(scope, input_names, output_names, container, operator_name=None, output_seq=0, **attrs):
    name = _create_name_or_use_existing_one(scope, 'RNN', operator_name)
    if container.target_opset <= 6:
        attrs['output_sequence'] = 1 if output_seq else 0
        op_version = 1
    elif container.target_opset <= 13:
        op_version = 7
    else:
        op_version = 14
    container.add_node('RNN', input_names, output_names, name=name, op_version=op_version, **attrs)


def apply_shape(scope, input_name, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Shape', operator_name)
    container.add_node('Shape', input_name, output_name, name=name, op_version=1)


def apply_sigmoid(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Sigmoid', input_name, output_name, container, operator_name)


def apply_softsign(scope, input_name, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Softsign', operator_name)
    container.add_node('Softsign', input_name, output_name, name=name, op_version=1)


# See alpha and gamma at https://github.com/keras-team/keras/blob/master/keras/activations.py#L80-L81
def apply_selu(scope, input_name, output_name, container, operator_name=None, alpha=1.673263, gamma=1.050701):
    _apply_unary_operation(scope, 'Selu', input_name, output_name, container, operator_name, alpha=alpha, gamma=gamma)


def apply_softmax(scope, input_name, output_name, container, operator_name=None, axis=None):
    name = _create_name_or_use_existing_one(scope, 'Softmax', operator_name)
    if axis is None:
        axis = 1 if container.target_opset < 13 else -1
    container.add_node('Softmax', input_name, output_name, name=name, axis=axis)


def apply_scaled_tanh(scope, input_name, output_name, container, operator_name=None, alpha=None, beta=None):
    if alpha is None:
        alpha = [1.0]
    if beta is None:
        beta = [1.0]
    if len(alpha) != 1 or len(beta) != 1:
        raise ValueError('alpha and beta must be 1-element lists')

    name = _create_name_or_use_existing_one(scope, 'ScaledTanh', operator_name)
    if container.target_opset < 9:
        attrs = {'name': name, 'alpha': alpha[0], 'beta': beta[0]}
        container.add_node('ScaledTanh', input_name, output_name, **attrs)
    else:
        # Define scalar a, initialize with parameter alpha.
        aName = scope.get_unique_variable_name(name + '_alpha')
        aShape = [len(alpha)] if len(alpha) == 1 else [len(alpha), 1, 1]
        container.add_initializer(aName, onnx_proto.TensorProto.FLOAT, aShape, alpha)

        # Define scalar b, initialize with parameter beta.
        bShape = [len(beta)] if len(beta) == 1 else [len(beta), 1, 1]
        bName = scope.get_unique_variable_name(name + '_beta')
        container.add_initializer(bName, onnx_proto.TensorProto.FLOAT, bShape, beta)

        # c = b * x
        cName = scope.get_unique_variable_name(name + '_c')
        apply_mul(scope, [input_name, bName], cName, container)

        # d = tanh(c)
        dName = scope.get_unique_variable_name(name + '_d')
        apply_tanh(scope, cName, dName, container)

        # output = a * d
        apply_mul(scope, [aName, dName], output_name, container)


def apply_slice(scope, input_name, output_name, container, starts, ends,
                axes=None, steps=None, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Slice', operator_name)

    if container.target_opset < 10:
        if axes is None:
            container.add_node('Slice', input_name, output_name, name=name,
                               starts=starts, ends=ends, op_version=1)
        else:
            container.add_node('Slice', input_name, output_name, name=name,
                               starts=starts, ends=ends, axes=axes, op_version=1)
    else:
        if container.target_opset == 10:
            op_version = 10
        else:
            op_version = 11
        inputs = input_name if isinstance(input_name, list) else [input_name]
        if isinstance(starts, str):
            starts_name = starts
        else:
            starts_name = scope.get_unique_variable_name('starts')
            container.add_initializer(starts_name, onnx_proto.TensorProto.INT64,
                                      [len(starts)], starts)

        if isinstance(ends, str):
            ends_name = ends
        else:
            ends_name = scope.get_unique_variable_name('ends')
            container.add_initializer(ends_name, onnx_proto.TensorProto.INT64,
                                      [len(ends)], ends)

        inputs.append(starts_name)
        inputs.append(ends_name)
        if axes:
            if isinstance(axes, str):
                axes_name = axes
            else:
                axes_name = scope.get_unique_variable_name('axes')
                container.add_initializer(axes_name, onnx_proto.TensorProto.INT64,
                                          [len(axes)], axes)
            inputs.append(axes_name)
        if steps:
            if not axes:
                inputs.append('')
            if isinstance(steps, str):
                steps_name = steps
            else:
                steps_name = scope.get_unique_variable_name('steps')
                container.add_initializer(steps_name, onnx_proto.TensorProto.INT64,
                                          [len(steps)], steps)
            inputs.append(steps_name)
        container.add_node('Slice', inputs, output_name, name=name,
                           op_version=op_version)


def apply_slice2(scope, input_name, output_name, container,
                 operator_name=None, starts=None, ends=None, axes=None, steps=None):
    assert starts is not None, 'the starts in slice op cannot be None'
    assert ends is not None, 'the ends in slice op cannot be None'
    return apply_slice(scope, input_name, output_name, container, starts, ends,
                       axes, steps, operator_name)


def apply_split(scope, input_name, output_names, container, operator_name=None, split=None, axis=0):
    name = _create_name_or_use_existing_one(scope, 'Split', operator_name)
    if container.target_opset <= 1:
        op_version = 1
    elif container.target_opset < 11:
        op_version = 2
    elif container.target_opset < 13:
        op_version = 11
    else:
        op_version = 13

    attrs = {'name': name}
    if split is not None:
        if container.target_opset < 13:
            attrs['split'] = split
        else:
            if not isinstance(input_name, list):
                input_name = [input_name]
            if isinstance(split, str):
                split_name = split
            else:
                split_name = scope.get_unique_variable_name(name + '_split')
                container.add_initializer(split_name, onnx_proto.TensorProto.INT64, [len(split)], split)
            input_name = input_name + [split_name]

    if axis is not None:
        attrs['axis'] = axis

    container.add_node('Split', input_name, output_names, op_version=op_version, **attrs)


def apply_sqrt(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Sqrt', input_name, output_name, container, operator_name=operator_name)


def _apply_squeeze_unsqueeze(scope, input_name, output_name, container, squeeze_str, operator_name=None, axes=None,
                             rank=0):
    name = _create_name_or_use_existing_one(scope, squeeze_str, operator_name)
    if container.target_opset < 13:
        if container.target_opset < 11:
            op_version = 1
            axes = [axis if axis >= 0 else axis + rank for axis in axes]
        else:
            op_version = 11
        container.add_node(squeeze_str, input_name, output_name, name=name, op_version=op_version, axes=axes)
    else:
        op_version = 13
        if not isinstance(input_name, list):
            input_name = [input_name]
        if isinstance(axes, str):
            container.add_node(squeeze_str, input_name + [axes], output_name, op_version=op_version, name=name)
        elif len(axes) == 0:
            container.add_node(squeeze_str, input_name, output_name, op_version=op_version, name=name)
        else:
            axes_name = scope.get_unique_variable_name(name + '_axes')
            container.add_initializer(axes_name, onnx_proto.TensorProto.INT64, [len(axes)], axes)
            container.add_node(squeeze_str, input_name + [axes_name], output_name, op_version=op_version, name=name)


def apply_squeeze(scope, input_name, output_name, container, operator_name=None, axes=None, rank=0):
    if axes is None:
        axes = []
    _apply_squeeze_unsqueeze(scope, input_name, output_name, container, 'Squeeze', operator_name, axes, rank)


def apply_sub(scope, input_names, output_name, container, operator_name=None, axis=None, broadcast=0):
    _apply_basic_numerical_operation(scope, 'Sub', input_names, output_name, container, operator_name=operator_name,
                                     axis=axis, broadcast=broadcast)


def apply_sum(scope, input_names, output_name, container, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'Sum', operator_name)
    if container.target_opset < 6:
        op_version = 1
    else:
        op_version = 6
    container.add_node('Sum', input_names, output_name, op_version=op_version, name=name)


def apply_tanh(scope, input_name, output_name, container, operator_name=None):
    _apply_unary_operation(scope, 'Tanh', input_name, output_name, container, operator_name)


def apply_thresholded_relu(scope, input_name, output_name, container, operator_name=None, alpha=None):
    if alpha is None:
        alpha = [1.0]

    name = _create_name_or_use_existing_one(scope, 'ThresholdedRelu', operator_name)
    attrs = {'name': name, 'alpha': alpha[0]}
    if container.target_opset < 10:
        # ThresholdedRelu graduated from an experimental op to a full op in opset 10
        # onnxruntime maintains support in the ONNX domain for ThresholdedRelu as a contrib op
        attrs['op_domain'] = "ai.onnx"
        op_version = 1
    else:
        op_version = 10
    container.add_node('ThresholdedRelu', input_name, output_name, op_version=op_version, **attrs)


def apply_tile(scope, input_name, output_name, container, operator_name=None, repeats=None):
    name = _create_name_or_use_existing_one(scope, 'Tile', operator_name)

    if repeats is None or (not isinstance(repeats, str) and all(repeat_count == 1 for repeat_count in repeats)):
        container.add_node('Identity', input_name, output_name, name=name)
        return

    if container.target_opset < 6:
        intermediate_input_name = input_name
        intermediate_output_name = None
        if isinstance(repeats, str):
            raise ValueError('repeats cannot be string type before opset 6')

        for axis, repeat_count in enumerate(repeats):
            if repeat_count == 1:
                continue

            # Create the 2nd input of Tile
            tile_tensor_name = scope.get_unique_variable_name(name + '_tile')
            container.add_initializer(tile_tensor_name, onnx_proto.TensorProto.FLOAT, [1], [float(repeat_count)])

            # Create the 3rd input of Tile
            axis_tensor_name = scope.get_unique_variable_name(name + '_axis')
            container.add_initializer(axis_tensor_name, onnx_proto.TensorProto.FLOAT, [1], [float(axis)])

            # Create tile for duplicating along one axis. After ONNX-1.2, we can duplicate along multiple axes, so we
            # don't have to iterate through all axes.
            intermediate_output_name = scope.get_unique_variable_name(name + '_input')
            container.add_node('Tile', [intermediate_input_name, tile_tensor_name, axis_tensor_name],
                               intermediate_output_name, name=name)

            # Use the output produced by this round as the input in the next iteration
            intermediate_input_name = intermediate_output_name

            # Create a new name for next Tile
            name = scope.get_unique_operator_name('Tile')

        # Use the last Tile name for the name of an Identity
        container.add_node('Identity', intermediate_output_name, output_name, op_version=1, name=name)
    else:
        # ONNX-1.2 has a new Tile and we use it here
        if isinstance(repeats, str):
            container.add_node('Tile', input_name + [repeats], output_name, op_version=6, name=name)
        else:
            repeat_tensor_name = scope.get_unique_variable_name(name + '_repeats')
            container.add_initializer(repeat_tensor_name, onnx_proto.TensorProto.INT64, [len(repeats)], repeats)
            container.add_node('Tile', [input_name, repeat_tensor_name], output_name, op_version=6, name=name)


def apply_topk(scope, input_name, output_names, container, k, operator_name=None):
    name = _create_name_or_use_existing_one(scope, 'TopK', operator_name)

    if container.target_opset < 10:
        if isinstance(k, str):
            raise ValueError('topk k cannot be string type before opset 10')
        container.add_node('TopK', input_name, output_names, name=name, k=k, op_version=1)
    else:
        if container.target_opset == 10:
            op_version = 10
        else:
            op_version = 11

        if isinstance(k, str):
            k_value_name = k
        else:
            k_value_name = scope.get_unique_variable_name('k_value')
            container.add_initializer(k_value_name, onnx_proto.TensorProto.INT64, [1], [k])
        container.add_node('TopK', input_name + [k_value_name], output_names, name=name, op_version=op_version)


def apply_transpose(scope, input_name, output_name, container, operator_name=None, perm=None):
    name = _create_name_or_use_existing_one(scope, 'Transpose', operator_name)
    container.add_node('Transpose', input_name, output_name, name=name, perm=perm)


def apply_upsample(scope, input_name, output_name, container, operator_name=None, mode='nearest',
                   coordinate_transformation_mode='asymmetric', scales=None):
    '''
    :param mode: nearest or linear
    :param scales: an integer list of scaling-up rate of all input dimensions
    '''
    if container.target_opset < 10:
        name = _create_name_or_use_existing_one(scope, 'Upsample', operator_name)
        inputs = [input_name]
        attrs = {'name': name}
        if container.target_opset < 7:
            if len(scales) != 4:
                raise ValueError('Need to specify a 4-element list the the scales of N-, C-, H-, and W-axes')
            attrs['height_scale'] = float(scales[2])
            attrs['width_scale'] = float(scales[3])
            attrs['mode'] = mode.upper()
            op_version = 1
        else:
            attrs['mode'] = mode.lower()
            if container.target_opset < 9:
                attrs['scales'] = list(map(float, scales))
                op_version = 7
            else:
                # scales moved from attribute to input in opset 9
                scales_tensor_name = scope.get_unique_variable_name(name + '_scales')
                container.add_initializer(scales_tensor_name, onnx_proto.TensorProto.FLOAT, [len(scales)], scales)
                inputs = [input_name, scales_tensor_name]
                op_version = 9

        container.add_node('Upsample', inputs, output_name, op_version=op_version, **attrs)
    else:
        # Upsample op is deprecated in ONNX opset 10
        # We implement Upsample through Resize instead
        apply_resize(scope, input_name, output_name, container, operator_name, mode, coordinate_transformation_mode,
                     scales)


def apply_unsqueeze(scope, input_name, output_name, container, operator_name=None, axes=None, rank=0):
    if axes is None:
        axes = [0]
    _apply_squeeze_unsqueeze(scope, input_name, output_name, container, 'Unsqueeze', operator_name, axes, rank)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

"""
The entry point to onnxconverter-common.
This framework performs optimization for ONNX models and
includes common utilities for ONNX converters.
"""
__version__ = "1.14.0"
__author__ = "Microsoft"
__producer__ = "OnnxMLTools"
__producer_version__ = __version__
__domain__ = "onnxconverter-common"
__model_version__ = 0

from .data_types import *  # noqa F403
from .onnx_ops import *  # noqa F403
from .container import *  # noqa F403
from .registration import *  # noqa F403
from .topology import *  # noqa F403
from .interface import *  # noqa F403
from .shape_calculator import *  # noqa F403
from .tree_ensemble import *  # noqa F403
from .utils import *  # noqa F403
from .case_insensitive_dict import *  # noqa F403
from .metadata_props import add_metadata_props, set_denotation
from .float16 import convert_tensor_float_to_float16, convert_float_to_float16, convert_float_to_float16_model_path
from .optimizer import optimize_onnx, optimize_onnx_graph, optimize_onnx_model
from .auto_mixed_precision import auto_convert_mixed_precision
from .auto_mixed_precision_model_path import auto_convert_mixed_precision_model_path



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

"""
This file defines the interface of the converter internal object for callback,
So the usage of the methods and properties list here will not be affected among the different versions.
"""

import abc


class ModelContainer:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def add_initializer(self, name, onnx_type, shape, content):
        """
        Add a TensorProto into the initializer list of the final ONNX model

        :param name: Variable name in the produced ONNX model.
        :param onnx_type: Element types allowed in ONNX tensor, e.g., TensorProto.FLOAT and TensorProto.STRING.
        :param shape: Tensor shape, a list of integers.
        :param content: Flattened tensor values (i.e., a float list or a float array).
        """
        return

    @abc.abstractmethod
    def add_node(self, op_type, inputs, outputs, op_domain='', op_version=1, **attrs):
        """
        Add a NodeProto into the node list of the final ONNX model. If the input operator's domain-version information
        cannot be found in our domain-version pool (a Python set), we may add it.

        :param op_type: A string (e.g., Pool and Conv) indicating the type of the NodeProto
        :param inputs: A list of strings. They are the input variables' names of the considered NodeProto
        :param outputs: A list of strings. They are the output variables' names of the considered NodeProto
        :param op_domain: The domain name (e.g., ai.onnx.ml) of the operator we are trying to add.
        :param op_version: The version number (e.g., 0 and 1) of the operator we are trying to add.
        :param attrs: A Python dictionary. Keys and values are attributes' names and attributes' values, respectively.
        """
        return


class OperatorBase:
    __metaclass__ = abc.ABCMeta

    @property
    @abc.abstractmethod
    def full_name(self):
        """
        Return a globally unique operator ID
        """
        pass

    @property
    @abc.abstractmethod
    def input_full_names(self):
        """
        Return all input variables' names
        """
        pass

    @property
    @abc.abstractmethod
    def output_full_names(self):
        """
        Return all outpu variables' names
        """
        pass

    @property
    @abc.abstractmethod
    def original_operator(self):
        """
        Return the original operator/layer
        """
        pass


class ScopeBase:
    __metaclass__ = abc.ABCMeta

    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

import numbers
import numpy as np
import warnings
import packaging.version as pv


def sparkml_installed():
    """
    Checks that *spark* is available.
    """
    try:
        import pyspark  # noqa F401
        return True
    except ImportError:
        return False


def sklearn_installed():
    """
    Checks that *scikit-learn* is available.
    """
    try:
        import sklearn  # noqa F401
        return True
    except ImportError:
        return False


def skl2onnx_installed():
    """
    Checks that *skl2onnx* converter is available.
    """
    try:
        import skl2onnx  # noqa F401
        return True
    except ImportError:
        return False


def coreml_installed():
    """
    Checks that *coremltools* is available.
    """
    try:
        import coremltools  # noqa F401
        return True
    except ImportError:
        return False


def keras2onnx_installed():
    """
    Checks that *keras2onnx* is available.
    """
    try:
        import keras2onnx  # noqa F401
        return True
    except ImportError:
        return False


def torch_installed():
    """
    Checks that *pytorch* is available.
    """
    try:
        import torch  # noqa F401
        return True
    except ImportError:
        return False


def caffe2_installed():
    """
    Checks that *caffe* is available.
    """
    try:
        import caffe2  # noqa F401
        return True
    except ImportError:
        return False


def libsvm_installed():
    """
    Checks that *libsvm* is available.
    """
    try:
        import svm  # noqa F401
        import svmutil  # noqa F401
        return True
    except ImportError:
        return False


def lightgbm_installed():
    """
    Checks that *lightgbm* is available.
    """
    try:
        import lightgbm  # noqa F401
        return True
    except ImportError:
        return False


def xgboost_installed():
    """
    Checks that *xgboost* is available.
    """
    try:
        import xgboost  # noqa F401
    except ImportError:
        return False
    from xgboost.core import _LIB
    try:
        _LIB.XGBoosterDumpModelEx
    except AttributeError:
        # The version is not recent enough even though it is version 0.6.
        # You need to install xgboost from github and not from pypi.
        return False
    from xgboost import __version__
    vers = pv.Version(__version__)
    allowed = pv.Version('0.7')
    if vers < allowed:
        warnings.warn('The converter works for xgboost >= 0.7. Earlier versions might not.')
    return True


def h2o_installed():
    """
    Checks that *h2o* is available.
    """
    try:
        import h2o  # noqa F401
    except ImportError:
        return False
    return True


def hummingbird_installed():
    """
    Checks that *Hummingbird* is available.
    """
    try:
        import hummingbird.ml  # noqa: F401

        return True
    except ImportError:
        return False


def get_producer():
    """
    Internal helper function to return the producer
    """
    from . import __producer__
    return __producer__


def get_producer_version():
    """
    Internal helper function to return the producer version
    """
    from . import __producer_version__
    return __producer_version__


def get_domain():
    """
    Internal helper function to return the model domain
    """
    from . import __domain__
    return __domain__


def get_model_version():
    """
    Internal helper function to return the model version
    """
    from . import __model_version__
    return __model_version__


def is_numeric_type(item):
    numeric_types = (int, float, complex)
    types = numeric_types

    if isinstance(item, list):
        return all(isinstance(i, types) for i in item)
    if isinstance(item, np.ndarray):
        return np.issubdtype(item.dtype, np.number)
    return isinstance(item, types)


def is_string_type(item):
    if isinstance(item, list):
        return all(isinstance(i, str) for i in item)
    if isinstance(item, np.ndarray):
        return np.issubdtype(item.dtype, np.str_)
    return isinstance(item, str)


def cast_list(type, items):
    return [type(item) for item in items]


def convert_to_python_value(var):
    if isinstance(var, numbers.Integral):
        return int(var)
    elif isinstance(var, numbers.Real):
        return float(var)
    elif isinstance(var, str):
        return str(var)
    else:
        raise TypeError('Unable to convert {0} to python type'.format(type(var)))


def convert_to_python_default_value(var):
    if isinstance(var, numbers.Integral):
        return int()
    elif isinstance(var, numbers.Real):
        return float()
    elif isinstance(var, str):
        return str()
    else:
        raise TypeError('Unable to find default python value for type {0}'.format(type(var)))


def convert_to_list(var):
    if isinstance(var, numbers.Real) or isinstance(var, str):
        return [convert_to_python_value(var)]
    elif isinstance(var, np.ndarray) and len(var.shape) == 1:
        return [convert_to_python_value(v) for v in var]
    elif isinstance(var, list):
        flattened = []
        if all(isinstance(ele, np.ndarray) and len(ele.shape) == 1 for ele in var):
            max_classes = max([ele.shape[0] for ele in var])
            flattened_one = []
            for ele in var:
                for i in range(max_classes):
                    if i < ele.shape[0]:
                        flattened_one.append(convert_to_python_value(ele[i]))
                    else:
                        flattened_one.append(convert_to_python_default_value(ele[0]))
            flattened += flattened_one
            return flattened
        elif all(isinstance(v, numbers.Real) or isinstance(v, str) for v in var):
            return [convert_to_python_value(v) for v in var]
        else:
            raise TypeError('Unable to flatten variable')
    else:
        raise TypeError('Unable to flatten variable')


def check_input_and_output_numbers(operator, input_count_range=None, output_count_range=None):
    '''
    Check if the number of input(s)/output(s) is correct

    :param operator: A Operator object
    :param input_count_range: A list of two integers or an integer. If it's a list the first/second element is the
    minimal/maximal number of inputs. If it's an integer, it is equivalent to specify that number twice in a list. For
    infinite ranges like 5 to infinity, you need to use [5, None].
    :param output_count_range: A list of two integers or an integer. See input_count_range for its format.
    '''
    if isinstance(input_count_range, list):
        min_input_count = input_count_range[0]
        max_input_count = input_count_range[1]
    elif isinstance(input_count_range, int) or input_count_range is None:
        min_input_count = input_count_range
        max_input_count = input_count_range
    else:
        raise RuntimeError('input_count_range must be a list or an integer')

    if isinstance(output_count_range, list):
        min_output_count = output_count_range[0]
        max_output_count = output_count_range[1]
    elif isinstance(output_count_range, int) or output_count_range is None:
        min_output_count = output_count_range
        max_output_count = output_count_range
    else:
        raise RuntimeError('output_count_range must be a list or an integer')

    if min_input_count is not None and len(operator.inputs) < min_input_count:
        raise RuntimeError(
            'For operator %s (type: %s), at least %s input(s) is(are) required but we got %s input(s) which are %s'
            % (operator.full_name, operator.type, min_input_count, len(operator.inputs), operator.input_full_names))

    if max_input_count is not None and len(operator.inputs) > max_input_count:
        raise RuntimeError(
            'For operator %s (type: %s), at most %s input(s) is(are) supported but we got %s input(s) which are %s'
            % (operator.full_name, operator.type, max_input_count, len(operator.inputs), operator.input_full_names))

    if min_output_count is not None and len(operator.outputs) < min_output_count:
        raise RuntimeError(
            'For operator %s (type: %s), at least %s output(s) is(are) produced but we got %s output(s) which are %s'
            % (operator.full_name, operator.type, min_output_count, len(operator.outputs), operator.output_full_names))

    if max_output_count is not None and len(operator.outputs) > max_output_count:
        raise RuntimeError(
            'For operator %s (type: %s), at most %s outputs(s) is(are) supported but we got %s output(s) which are %s'
            % (operator.full_name, operator.type, max_output_count, len(operator.outputs), operator.output_full_names))


def check_input_and_output_types(operator, good_input_types=None, good_output_types=None):
    '''
    Check if the type(s) of input(s)/output(s) is(are) correct

    :param operator: A Operator object
    :param good_input_types: A list of allowed input types (e.g., [FloatTensorType, Int64TensorType]) or None. None
    means that we skip the check of the input types.
    :param good_output_types: A list of allowed output types. See good_input_types for its format.
    '''
    if good_input_types is not None:
        for variable in operator.inputs:
            if type(variable.type) not in good_input_types:
                raise RuntimeError('Operator %s (type: %s) got an input %s with a wrong type %s. Only %s are allowed'
                                   % (operator.full_name, operator.type, variable.full_name, type(variable.type),
                                      good_input_types))

    if good_output_types is not None:
        for variable in operator.outputs:
            if type(variable.type) not in good_output_types:
                raise RuntimeError('Operator %s (type: %s) got an output %s with a wrong type %s. Only %s are allowed'
                                   % (operator.full_name, operator.type, variable.full_name, type(variable.type),
                                      good_output_types))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###########################################################################

"""
Script for parsing profiling results from onnxruntime. Outputs to console or CSV.
Prints total durations grouped by node (-n), optype (-t), or step (-s)
Script is standalone and can be downloaded and run as prefstats.py

Ex:
    python -m onnxconvert_common.perfstats -t -l 5 trace.json   # List durations by optype
    (or python perfstats.py -t -l 5 trace.json)

Result:
    op_type               duration    percent     count
    --------------------  ----------  ----------  ----------
    MemcpyFromHost        3388472     66.0656981  193
    Conv                  1205958     23.5127978  5
    Loop                  451989      8.81251751  5
    Add                   23115       0.45067765  3482
    Concat                14228       0.27740608  323

Get a trace file from onnxruntime first with:

    import onnxruntime as ort
    sess_options = ort.SessionOptions()
    sess_options.enable_profiling = True
    sess_profile = ort.InferenceSession(onnx_model_str, sess_options)
    ...
    sess.run(None, feed_dict)
    prof_file = sess_profile.end_profiling()
    print(prof_file)

If you see non-onnx ops like MemcpyFromHost, these are inserted into the graph by ORT
Get a new graph with all inserted ops with:

    sess_options.graph_optimization_level = rt.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
    sess_options.optimized_model_filepath = "optimized_graph.onnx"
"""

import json
import argparse
import shutil
import csv
from collections import namedtuple, defaultdict

_HELP_TEXT = """
Usage Examples:

python -m onnxconvert_common.perfstats -t trace.json   # List results by type
python -m onnxconvert_common.perfstats -n -l 10 trace.json   # List top 10 nodes
python -m onnxconvert_common.perfstats -n trace.json -q t=Conv   # List conv nodes
python -m onnxconvert_common.perfstats -n trace.json -q t!=Conv   # List non-conv nodes
python -m onnxconvert_common.perfstats -n trace.json -q t=Conv;n!=NAME   # List conv nodes except NAME
"""


raw_entry_headers = \
    ["name", "duration", "op_type", "provider", "graph_index", "parameter_size", "activation_size", "output_size"]
raw_entry_col_widths = [None, 10, 20, 20, 11, 14, 15, 11]
RawEntry = namedtuple("RawEntry", raw_entry_headers)
node_entry_headers = ["name", "duration", "op_type", "provider", "percent", "count"]
node_entry_col_widths = [None, 10, 20, 20, 10, 10]
NodeEntry = namedtuple("NodeEntry", node_entry_headers)
step_entry_headers = ["name", "duration", "op_type", "provider", "percent"]
step_entry_col_widths = [None, 10, 20, 20, 10]
StepEntry = namedtuple("StepEntry", step_entry_headers)
type_entry_headers = ["op_type", "duration", "percent", "count"]
type_entry_col_widths = [20, 10, 10, 10]
OpTypeEntry = namedtuple("OpTypeEntry", type_entry_headers)


def compute_step_entries(raw_entries):
    total_duration = sum(entry.duration for entry in raw_entries)
    step_entries = []
    for entry in raw_entries:
        percent = entry.duration * 100 / total_duration
        step_entries.append(StepEntry(entry.name, entry.duration, entry.op_type, entry.provider, percent))
    step_entries.sort(key=lambda x: -x.duration)
    return step_entries


def compute_node_entries(raw_entries):
    name_to_data = defaultdict(list)
    total_duration = sum(entry.duration for entry in raw_entries)
    for entry in raw_entries:
        name_to_data[entry.name].append(entry)
    node_entries = []
    for name, entries in name_to_data.items():
        duration = sum(entry.duration for entry in entries)
        percent = duration * 100 / total_duration
        op_type = entries[0].op_type
        provider = entries[0].provider
        node_entries.append(NodeEntry(name, duration, op_type, provider, percent, len(entries)))
    node_entries.sort(key=lambda x: -x.duration)
    return node_entries


def compute_op_type_entries(raw_entries):
    type_to_data = defaultdict(list)
    total_duration = sum(entry.duration for entry in raw_entries)
    for entry in raw_entries:
        type_to_data[entry.op_type].append(entry)
    op_type_entries = []
    for op_type, entries in type_to_data.items():
        duration = sum(entry.duration for entry in entries)
        percent = duration * 100 / total_duration
        op_type_entries.append(OpTypeEntry(op_type, duration, percent, len(entries)))
    op_type_entries.sort(key=lambda x: -x.percent)
    return op_type_entries


def read_raw_entries(profile_path):
    with open(profile_path, "r") as f:
        data = json.load(f)
    if isinstance(data, dict):
        data = data['traceEvents']
    entries = []
    for item in data:
        cat = item.get("cat")
        if cat not in ["Node", "Op"]:
            continue
        arg = item.get('args')
        if not arg:
            continue
        provider = arg.get("provider")
        op = arg.get("op_name")
        if op:
            name = item['name']
            if not name.endswith("_kernel_time"):
                continue
            dur = item['dur']
            name = name.replace("_kernel_time", "")
            graph_index = arg.get('graph_index')
            parameter_size = arg.get('parameter_size')
            activation_size = arg.get('activation_size')
            output_size = arg.get('output_size')
        if not op:
            continue
        entries.append(RawEntry(name, dur, op, provider, graph_index, parameter_size, activation_size, output_size))
    return entries


def get_args():
    """Parse commandline."""
    parser = argparse.ArgumentParser(description="Parses a json profiling file from onnx runtime",
                                     formatter_class=argparse.RawDescriptionHelpFormatter, epilog=_HELP_TEXT)
    parser.add_argument("input", help=".json file from onnx runtime")
    parser.add_argument("-t", "--type", action="store_true", help="total execution time per op type (sorted)")
    parser.add_argument("-n", "--node", action="store_true", help="total execution time per node (sorted)")
    parser.add_argument("-s", "--step", action="store_true", help="times for each execution step (sorted)")
    parser.add_argument("-r", "--raw", action="store_true", help="unsorted raw data")
    parser.add_argument("-d", "--data-only", action="store_true", help="don't include headers")
    parser.add_argument("-q", "--query", help="only include entries satisfying the provided query")
    parser.add_argument("-l", "--limit", type=int, default=-1, help="only show first n results")
    parser.add_argument("-o", "--output", help="output to csv file")
    args = parser.parse_args()
    if sum(bool(a) for a in [args.type, args.node, args.step, args.raw]) != 1:
        print("exactly one of flags -t, -n, -s, -r must be provided")
        exit(1)
    try:
        if args.query:
            args.query = Query(args.query)
    except Exception:
        print("invalid query: %r" % args.query)
        exit(1)
    return args


class QueryClause:
    def __init__(self, clause_string):
        self.rule_type = 'inc'
        letter = None
        if '!=' in clause_string:
            self.rule_type = 'exc'
            letter, clause_string = clause_string.split('!=')
        elif '=' in clause_string:
            letter, clause_string = clause_string.split('=')
        self.match_name = letter in [None, 'n']
        self.match_type = letter in [None, 't']
        self.patterns = set(clause_string.split(','))

    def match(self, entry):
        if isinstance(entry, (NodeEntry, RawEntry)) and self.match_name and entry.name in self.patterns:
            return self.rule_type
        if self.match_type and entry.op_type in self.patterns:
            return self.rule_type
        return None


class Query:
    def __init__(self, query_string):
        self.clauses = [QueryClause(s) for s in query_string.split(";")]
        self.no_inc = not any(c.rule_type == 'inc' for c in self.clauses)

    def match(self, entry):
        matches = [c.match(entry) for c in self.clauses]
        return (self.no_inc or 'inc' in matches) and 'exc' not in matches


class TablePrinter:
    def __init__(self, col_widths, padding=2, min_width=5):
        self.col_widths = col_widths
        self.unknown_cnt = col_widths.count(None)
        self.padding = padding
        self.fixed_sum = sum(w for w in col_widths if w is not None) + self.padding * (len(col_widths) - 1)
        self.min_width = min_width

    def get_col_widths(self, total_width):
        remaining_width = total_width - self.fixed_sum
        computed_widths = []
        for i in range(self.unknown_cnt):
            w = remaining_width // (self.unknown_cnt - i)
            remaining_width -= w
            computed_widths.append(max(w, self.min_width))
        col_widths = []
        for w in self.col_widths:
            if w is None:
                col_widths.append(computed_widths.pop())
            else:
                col_widths.append(w)
        return col_widths

    def format(self, entry, width):
        if isinstance(entry, float):
            entry = ("%." + str(width) + "f") % entry
            return entry[:width]
        else:
            entry = str(entry)
        if len(entry) > width:
            x = (width - 3) // 2
            y = width - 3 - x
            entry = entry[:x] + "..." + entry[-y:]
        return entry + " " * (width - len(entry))

    def print_divider(self):
        total_width = shutil.get_terminal_size((80, 20)).columns
        col_widths = self.get_col_widths(total_width)
        line = (" " * self.padding).join("-" * w for w in col_widths)
        print(line)

    def print(self, entries):
        total_width = shutil.get_terminal_size((80, 20)).columns
        col_widths = self.get_col_widths(total_width)
        line = (" " * self.padding).join(self.format(e, w) for e, w in zip(entries, col_widths))
        print(line)


def main():
    args = get_args()
    entries = read_raw_entries(args.input)
    if args.type:
        entries = compute_op_type_entries(entries)
    elif args.node:
        entries = compute_node_entries(entries)
    elif args.step:
        entries = compute_step_entries(entries)

    exc_entries = []
    if args.query:
        exc_entries.extend(e for e in entries if not args.query.match(e))
        entries = [e for e in entries if args.query.match(e)]
    if args.limit >= 0:
        exc_entries.extend(entries[args.limit:])
        entries = entries[:args.limit]

    if args.type:
        col_widths = type_entry_col_widths
        headers = type_entry_headers
    elif args.node:
        col_widths = node_entry_col_widths
        headers = node_entry_headers
    elif args.step:
        col_widths = step_entry_col_widths
        headers = step_entry_headers
    elif args.raw:
        col_widths = raw_entry_col_widths
        headers = raw_entry_headers

    if args.output:
        with open(args.output, mode='w', newline='') as f:
            writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            if not args.data_only:
                writer.writerow(headers)
            for entry in entries:
                writer.writerow(entry)
    else:
        printer = TablePrinter(col_widths)
        if not args.data_only:
            printer.print(headers)
            printer.print_divider()
        for entry in entries:
            printer.print(entry)


if __name__ == '__main__':
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

import numpy as np
from onnx import numpy_helper, mapping, helper


class OnnxGraphContext:
    stopping_initializers = []

    def __init__(self, graph_proto, nodelist):
        self.initializers = {ts_.name: ts_ for ts_ in graph_proto.initializer}
        # self.nodes = {nd_.name for nd_ in graph_proto.node}
        self.tensor_to_node = {}
        for nd_ in nodelist:
            self.tensor_to_node.update({ts_: nd_ for ts_ in nd_.output})
        self.variables = {}

    def add_value_of_node(self, name, value):
        if name in self.variables:
            assert False, "The tensor({}) already was assigned!".format(name)
        else:
            self.variables[name] = value

    @staticmethod
    def get_attribute(node, attr_name, default_value=None):
        found = [attr for attr in node.attribute if attr.name == attr_name]
        if found:
            return helper.get_attribute_value(found[0])
        return default_value

    @staticmethod
    def get_attr_graph(node):
        if node.op_type in ['Loop', 'Scan']:
            inner_graph = OnnxGraphContext.get_attribute(node, 'body')
            return {'body': inner_graph}
        elif node.op_type in ['If']:
            else_graph = OnnxGraphContext.get_attribute(node, 'else_branch')
            then_graph = OnnxGraphContext.get_attribute(node, 'then_branch')
            return {'else_branch': else_graph, 'then_branch': then_graph}
        else:
            return {}

    def calculate(self, node):
        func_name = '_On' + node.op_type
        func = type(self).__dict__.get(func_name, None)
        if func is None:
            return None

        inputs = []
        for ts_ in node.input:
            if ts_ in self.initializers:
                inputs.append(numpy_helper.to_array(self.initializers[ts_]))
            elif ts_ in self.variables:
                inputs.append(self.variables[ts_])
            else:
                return None

        output_values = func(self, node, inputs)
        for idx_, ots_ in enumerate(node.output):
            self.add_value_of_node(ots_, output_values[idx_])

        return output_values

    def _OnIdentity(self, node, inputs):
        return inputs

    def _OnConstant(self, node, inputs):
        return [numpy_helper.to_array(OnnxGraphContext.get_attribute(node, 'value'))]

    def _OnAdd(self, node, inputs):
        return [np.add(inputs[0], inputs[1])]

    def _OnCast(self, node, inputs):
        np_dtype = mapping.TENSOR_TYPE_TO_NP_TYPE[OnnxGraphContext.get_attribute(node, 'to')]
        casted = inputs[0].astype(np_dtype)
        return [casted]

    def _OnGather(self, node, inputs):
        data_val = inputs[0]
        idx_val = inputs[1]
        axis = OnnxGraphContext.get_attribute(node, 'axis')
        retval = np.take(data_val, idx_val, axis).astype(inputs[0].dtype)
        return [retval]

    def _OnGreater(self, node, inputs):
        return [np.greater(inputs[0], inputs[1])]

    def _OnLess(self, node, inputs):
        return [np.less(inputs[0], inputs[1])]

    def _OnMul(self, node, inputs):
        return [np.multiply(inputs[0], inputs[1])]

    def _OnNot(self, node, inputs):
        return [np.logical_not(inputs[0])]

    def _OnTranspose(self, node, inputs):
        perm_attr = OnnxGraphContext.get_attribute(node, 'perm')
        retval = inputs[0].transpose(perm_attr)
        return [retval]

    def _OnRange(self, node, inputs):
        retval = np.arange(inputs[0].item(0),
                           inputs[1].item(0),
                           inputs[2].item(0),
                           dtype=inputs[0].dtype)
        return [retval]

    def _OnSlice(self, node, inputs):
        data_val = inputs[0]
        rank = len(inputs[0].shape)
        starts = inputs[1].tolist()
        ends = inputs[2].tolist()
        axes = inputs[3].tolist() if len(node.input) > 3 else list(range(0, rank))
        steps = inputs[4].tolist() if len(node.input) > 4 else [1] * rank
        exp_all = list()
        for axis_ in range(0, rank):
            if axis_ in axes:
                idx = axes.index(axis_)
                cur_starts = starts[idx]
                cur_ends = ends[idx]
                cur_step = steps[idx]
            else:
                cur_starts = 0
                cur_ends = np.iinfo(np.int64).max
                cur_step = 1

            cur_starts = max(cur_starts, -inputs[0].shape[axis_])
            cur_ends = min(cur_ends, inputs[0].shape[axis_])
            exp = slice(cur_starts, cur_ends, cur_step)
            exp_all.append(exp)

        retval = data_val[tuple(exp_all)]
        return [retval]

    def _OnDiv(self, node, inputs):
        return [np.divide(inputs[0], inputs[1])]

    def _OnPow(self, node, inputs):
        return [np.power(inputs[0], inputs[1])]

    def _OnSqrt(self, node, inputs):
        return [np.sqrt(inputs[0].tolist(), dtype=inputs[0].dtype)]

    def _OnSub(self, node, inputs):
        return [np.subtract(inputs[0], inputs[1])]

    def _OnUnsqueeze(self, node, inputs):
        axes = OnnxGraphContext.get_attribute(node, 'axes')
        if axes is None:
            axes = inputs[1]
        shape_in = inputs[0].shape
        dims_out = len(shape_in) + len(axes)
        shape_in = iter(shape_in)
        shape_out = [None] * dims_out
        for idx_ in axes:
            shape_out[idx_] = 1
        for ind, val in enumerate(shape_out):
            if val is None:
                shape_out[ind] = next(shape_in)

        retval = inputs[0].reshape(shape_out)
        return [retval]

    def _OnReshape(self, node, inputs):
        retval = inputs[0].reshape(inputs[1])
        return [retval]


def _fix_unamed_node(nodelist):
    node_id = [1]
    name_set = set(nd_.name if nd_.name else '' for nd_ in nodelist)

    def _ensure_node_named(node, incre_id):
        if node.name:
            return node
        while True:
            name = node.op_type.lower() + "_{}".format(incre_id[0])
            incre_id[0] += 1
            if name not in name_set:
                node.name = name
                break

        return node

    named_nodes = [_ensure_node_named(nd_, node_id) for nd_ in nodelist]
    return named_nodes


def reserve_node_for_embedded_graph(nodelist):
    # type: (onnx.GraphProto)->(onnx.GraphProto, frozenset)
    nodelist = _fix_unamed_node(nodelist)
    ginputs = []
    for nd_ in nodelist:
        for _, subgraph_ in OnnxGraphContext.get_attr_graph(nd_).items():
            inner_inputs = frozenset([i_.name for i_ in subgraph_.input])
            for sub_nd_ in subgraph_.node:
                ginputs.extend([i_ for i_ in sub_nd_.input if i_ not in inner_inputs])
    ginputs.extend(OnnxGraphContext.stopping_initializers)
    return nodelist, frozenset(ginputs)


def _dfs_calc(graph, node, reserved_names, node_status):
    # type: (OnnxGraphContext, onnx.NodeProto, frozenset, dict) -> int
    if node.name in node_status:
        return node_status[node.name]

    if len(node.input) == 0:
        assert node.op_type in ['Constant', 'RandomNormal', 'RandomUniform'], \
            "Assume only the generator operation node hasn't any inputs"
        status = -1
        if node.op_type == 'Constant':
            graph.calculate(node)
            node_status[node.name] = 0
        return status
    else:
        calc_status = [0] * len(node.input)
        for idx_, ts_ in enumerate(node.input):
            if ts_ in graph.initializers or ts_ in graph.variables:
                calc_status[idx_] = -1 if ts_ in reserved_names else 0
            elif ts_ not in graph.tensor_to_node:  # input of graph
                calc_status[idx_] = -1
            else:
                calc_status[idx_] = _dfs_calc(graph, graph.tensor_to_node[ts_], reserved_names, node_status)

        status_up = max(calc_status)
        status_low = min(calc_status)
        if status_low < 0:
            status = - max(-status_low, abs(status_up)) - 1
        else:
            status = status_up + 1

        node_status[node.name] = status
        if status > 0:
            if any(o_ in reserved_names for o_ in node.output):
                status = -status
            else:
                graph.calculate(node)
        return status


def _is_initializer_existed(intlzer, initializers):
    val = numpy_helper.to_array(intlzer)
    for i_ in initializers:
        if intlzer.name == i_.name:
            if np.allclose(val, numpy_helper.to_array(i_)):
                return True

    return False


def _remove_unused_initializers(nodes, initializers, reversed_names, outer_initializers=None):
    nodes_input_set = set()
    for nd_ in nodes:
        nodes_input_set.update(n_ for n_ in nd_.input)

    lst = [intlz_ for intlz_ in initializers if intlz_.name in nodes_input_set or intlz_.name in reversed_names]
    if outer_initializers is not None:
        return [intlz_ for intlz_ in lst if not _is_initializer_existed(intlz_, outer_initializers)]
    return lst


def const_folding_optimizer(graph, outer_graph=None):
    # type: (onnx.GraphProto, onnx.GraphProto)->onnx.GraphProto
    nodelist, reserved_names = reserve_node_for_embedded_graph(graph.node)
    opt_graph = OnnxGraphContext(graph, nodelist)
    node_status = {}
    for ts_ in graph.output:
        _dfs_calc(opt_graph, opt_graph.tensor_to_node[ts_.name], reserved_names, node_status)

    graph.initializer.extend([numpy_helper.from_array(ts_, nm_) for nm_, ts_ in opt_graph.variables.items()])
    new_nodes = [nd_ for nd_ in nodelist if nd_.name in node_status]
    new_nodes = [nd_ for nd_ in new_nodes if nd_.output[0] not in opt_graph.variables]

    def node_key(nd_):
        return abs(node_status[nd_.name])

    new_nodes.sort(key=node_key)
    pruned_initilizers = _remove_unused_initializers(new_nodes, graph.initializer, reserved_names,
                                                     None if outer_graph is None else outer_graph.initializer)
    del graph.node[:]
    graph.node.extend(new_nodes)
    del graph.initializer[:]
    graph.initializer.extend(pruned_initilizers)

    for nd_ in graph.node:
        for aname_, subgraph_ in OnnxGraphContext.get_attr_graph(nd_).items():
            opt_inner_graph = const_folding_optimizer(subgraph_, graph)
            lst_attrs = list(nd_.attribute)
            del nd_.attribute[:]
            lst_attrs = [helper.make_attribute(aname_, opt_inner_graph) if
                         attr.name == aname_ else attr for attr in lst_attrs]
            nd_.attribute.extend(lst_attrs)

    return graph



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################

# This dictionary defines the converters which can be invoked in the conversion framework defined in topology.py. A key
# in this dictionary is an operator's unique ID (e.g., string and type) while the associated value is the callable
# object used to convert the operator specified by the key.
_converter_pool = {}

# This dictionary defines the shape calculators which can be invoked in the conversion framework defined in
# topology.py. A key in this dictionary is an operator's unique ID (e.g., string and type) while the associated value
# is the callable object used to infer the output shape(s) for the operator specified by the key.
_shape_calculator_pool = {}


def register_converter(operator_name, conversion_function, overwrite=False):
    '''
    :param operator_name: A unique operator ID. It is usually a string but you can use a type as well
    :param conversion_function: A callable object
    :param overwrite: By default, we raise an exception if the caller of this function is trying to assign an existing
    key (i.e., operator_name) a new value (i.e., conversion_function). Set this flag to True to enable overwriting.
    '''
    if not overwrite and operator_name in _converter_pool:
        raise ValueError('We do not overwrite registrated converter by default')
    _converter_pool[operator_name] = conversion_function


def get_converter(operator_name):
    '''
    Given an Operator object (named operator) defined in topology.py, we can retrieve its conversion function.
    >>> from onnxmltools.convert.common._topology import Operator
    >>> operator = Operator('dummy_name', 'dummy_scope', 'dummy_operator_type', None)
    >>> get_converter(operator.type)  # Use 'dummy_operator_type' for dictionary looking-up

    :param operator_name: An operator ID
    :return: a conversion function for a specific Operator object
    '''
    if operator_name not in _converter_pool:
        raise ValueError('Unsupported conversion for operator %s' % operator_name)
    return _converter_pool[operator_name]


def register_shape_calculator(operator_name, calculator_function, overwrite=False):
    '''
    :param operator_name: A unique operator ID. It is usually a string but you can use a type as well
    :param calculator_function: A callable object
    :param overwrite:  By default, we raise an exception if the caller of this function is trying to assign an existing
    key (i.e., operator_name) a new value (i.e., calculator_function). Set this flag to True to enable overwriting.
    '''
    if not overwrite and operator_name in _shape_calculator_pool:
        raise ValueError('We do not overwrite registrated shape calculator by default')
    _shape_calculator_pool[operator_name] = calculator_function


def get_shape_calculator(operator_name):
    '''
    Given an Operator object (named operator) defined in topology.py, we can retrieve its shape calculation function.
    >>> from onnxmltools.convert.common._topology import Operator
    >>> operator = Operator('dummy_name', 'dummy_scope', 'dummy_operator_type', None)
    >>> get_shape_calculator(operator.type)  # Use 'dummy_operator_type' for dictionary looking-up

    :param operator_name: An operator ID
    :return: a shape calculation function for a specific Operator object
    '''
    if operator_name not in _shape_calculator_pool:
        raise ValueError('Unsupported shape calculation for operator %s' % operator_name)
    return _shape_calculator_pool[operator_name]



# -*- coding: utf-8 -*-
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
###############################################################################
import copy
import onnx
import logging
import numpy as np
from onnx import onnx_pb as onnx_proto
from inspect import getfullargspec

from . import onnx_ops
from .registration import register_converter
from .topology import Topology, convert_topology
from .oopb import OnnxOperatorBuilder
from .onnx_ex import OPSET_TO_IR_VERSION, get_maximum_opset_supported
from .data_types import (DoubleTensorType, FloatTensorType,
                         Int64TensorType, Int32TensorType, BooleanTensorType,
                         Complex64TensorType, Complex128TensorType, StringTensorType)

_logger = logging.getLogger(__name__)


class GraphFunctionType:
    D = DoubleTensorType
    F = FloatTensorType
    I = Int64TensorType  # noqa: E741 ambiguous variable name 'I'
    B = BooleanTensorType
    I32 = Int32TensorType
    C64 = Complex64TensorType
    C128 = Complex128TensorType
    S = StringTensorType

    d = DoubleTensorType(shape=[])
    f = FloatTensorType(shape=[])
    i = Int64TensorType(shape=[])
    i32 = Int32TensorType(shape=[])
    b = BooleanTensorType(shape=[])
    c64 = Complex64TensorType(shape=[])
    c128 = Complex64TensorType(shape=[])
    s = StringTensorType(shape=[])


def _get_python_function_arguments(f):
    """
    Helper to get the parameter names and annotations of a Python function.
    """
    # Note that we only return non-optional arguments (we assume that any optional args are not specified).
    # This allows to, e.g., accept max(a, b, *more, name='') as a binary function
    param_specs = getfullargspec(f)
    annotations = param_specs.annotations
    arg_names = param_specs.args
    # "if this tuple has n elements, they correspond to the last n elements listed in args"
    defaults = param_specs.defaults
    if defaults:
        # we allow Function(functions with default arguments),
        # but those args will always have default values since CNTK Functions do not support this
        arg_names = arg_names[:-len(defaults)]
    return arg_names, annotations


class _SimpleRawModelContainer(object):
    def __init__(self, inputs, outputs):
        self.input_names = inputs
        self.output_names = outputs


class Graph:
    """
    The ONNX Graph build from a python function, or load it from an ONNX object
    """
    opset = get_maximum_opset_supported()

    inference_runtime = None

    def __init__(self, name):
        self._name = name
        self._oxml = None
        self._onnxfunc = None
        self._onnxfunc_args = []
        self._onnxfunc_kwargs = {}

    @property
    def oxml(self):
        return self.to_model()

    @property
    def name(self):
        return self._name

    def to_model(self):
        if self._oxml is None:
            self._defer_building()
        return self._oxml

    def _bind(self, oxml, inputs, outputs):
        ox_graph = oxml.graph
        initializer_set = {
            initializer.name for initializer in ox_graph.initializer}
        model_inputs = [
            input.name for input in ox_graph.input if input.name not in initializer_set]
        model_outputs = [output.name for output in ox_graph.output]
        if inputs is None:
            inputs = model_inputs
        else:
            assert {input for input in inputs} == {
                input for input in
                model_inputs}, "User-specified set of inputs does not match actual set"
        if outputs is None:
            outputs = model_outputs
        else:
            assert {output for output in outputs} == {
                output for output in
                model_outputs}, "User-specified set of outputs does not match actual set"
        self._oxml = oxml
        self._inputs = inputs
        self._outputs = outputs

    @staticmethod
    def _create(func, *args, **kwargs):
        name = kwargs.get('name', None)
        f_name = name or func.__name__
        graph = Graph(f_name)
        graph._onnxfunc = func
        graph._onnxfunc_args = args
        graph._onnxfunc_kwargs = kwargs
        return graph

    @staticmethod
    def trace(*args, **kwargs):
        """
        This is the decorator. Example:
        @Graph.trace(outputs="logits")
        def model(source_sequence):
            ...
        """
        if len(args) > 0 and hasattr(args[0], '__call__'):  # first arg is function
            return Graph._create(args[0])
        else:
            return lambda f: Graph._create(f, *args, **kwargs)

    @staticmethod
    def _to_list(element_or_list):
        if element_or_list is None:
            return []

        return element_or_list if isinstance(
            element_or_list, (list, tuple)) else [element_or_list]

    @staticmethod
    def _on_conversion(scope, operator, container):
        with OnnxOperatorBuilderX(container, scope).as_default(operator.full_name) as ox:  # type: OnnxOperatorBuilderX
            f = operator.raw_operator
            # if ox.upper_context is not None:
            #     container.enable_optimizer = False  # optimizer on a subgraph is not supported yet.
            container.enable_optimizer = False
            fn_inputs = [ox.arg(arg_name) for arg_name in operator.input_full_names]
            f_outputs = f(*fn_inputs)
            outputs = operator.output_full_names
            if outputs:
                if isinstance(f_outputs, Tensor):
                    f_outputs = ox.identity([f_outputs], outputs=outputs)
                    n_outputs = 1
                else:
                    f_outputs = [ox.identity([f_output], outputs=[output_name])
                                 for f_output, output_name in zip(f_outputs, outputs)]
                    n_outputs = len(f_outputs)
            assert n_outputs == len(
                outputs), "Function {}() returned {} but {} were declared".format(
                operator.full_name, n_outputs, len(outputs))

    @staticmethod
    def _enforce_opset_version(oxml):
        for im_ in oxml.opset_import:
            if (im_.domain == '' or im_.domain == 'ai.onnx') and \
                    im_.version != Graph.opset:
                im_.version = Graph.opset
                opv = Graph.opset
                irv = OPSET_TO_IR_VERSION.get(opv, onnx_proto.IR_VERSION)
                oxml.ir_version = irv
                _logger.warning('The maximum opset needed by this model is updated to %d.' % Graph.opset)

    def _build_graph(self, f, input_types=None, output_types=None, outputs=None):
        input_types = Graph._to_list(input_types)
        output_types = Graph._to_list(output_types)
        outputs = Graph._to_list(outputs)
        if not outputs:
            outputs.append(f.__name__)

        f_name = self._name
        arg_names, _ = _get_python_function_arguments(f)
        raw_model = _SimpleRawModelContainer(arg_names, outputs)
        topo = Topology(raw_model)
        top_level = topo.declare_scope(f_name)
        graph_opname = f_name
        op_whole = top_level.declare_local_operator(graph_opname, f)
        register_converter(op_whole.type, type(self)._on_conversion, overwrite=True)
        for i_ in raw_model.input_names:
            vi_ = top_level.get_local_variable_or_declare_one(i_,
                                                              FloatTensorType(shape=[None]) if not input_types else
                                                              input_types[
                                                                  arg_names.index(i_)])
            op_whole.inputs.append(vi_)
        for o_ in raw_model.output_names:
            vo_ = top_level.get_local_variable_or_declare_one(o_,
                                                              FloatTensorType(shape=[None]) if not output_types else
                                                              output_types[outputs.index(o_)])
            op_whole.outputs.append(vo_)

        oxml = convert_topology(topo, f_name, "onnx.fn: {}".format(f_name), target_opset=Graph.opset)
        type(self)._enforce_opset_version(oxml)
        self._bind(oxml, arg_names, outputs)
        return self

    @staticmethod
    def _map_function_arguments(params, params_set, *args, **kwargs):
        """
        Helper to determine the argument map for use with various call operations.
        Returns a dictionary from parameters to whatever arguments are passed.
        Accepted are both positional and keyword arguments.
        This mimics Python's argument interpretation, except that keyword arguments are not optional.
        """
        # start with positional arguments
        arg_map = dict(zip(params, args))

        # now look up keyword arguments
        if len(kwargs) != 0:
            for name, arg in kwargs.items():  # keyword args are matched by name
                if name not in params_set:
                    raise TypeError("got an unexpected keyword argument '%s'" % name)
                if name in arg_map:
                    raise SyntaxError("got multiple values for argument '%s'" % name)
                arg_map[name] = arg  # add kw argument to dict
        assert len(arg_map) == len(params)

        return arg_map

    def _defer_building(self):
        if self._oxml is None:
            self._build_graph(self._onnxfunc, *self._onnxfunc_args, **self._onnxfunc_kwargs)

    def _argument_map(self, *args, **kwargs):
        """
        Determines the {placeholder: variable} map for use with various call operations
        Returns a dictionary from this function's placeholders to whatever arguments are passed.
        Accepted are both positional and keyword arguments.
        This mimics Python's argument interpretation, except that keyword arguments are not optional
        (there is no concept of default value).
        """
        params = self._inputs
        if len(args) + len(kwargs) != len(params):
            raise TypeError("Graph invocation expected {} arguments, got {}".format(
                len(params), len(args) + len(kwargs)))
        params_set = {arg for arg in params}
        return Graph._map_function_arguments(params, params_set, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        self._defer_building()
        # parse argument list and map to the function's input
        arg_map = self._argument_map(*args, **kwargs)
        # determine whether this is eval() or clone()
        is_symbolic = any(isinstance(arg, Tensor) for arg in arg_map.values())
        if is_symbolic:
            first_arg = next(iter(arg_map.values()))
            ox = first_arg.ox
            output_map = {output: None for output in self._outputs}
            # @TODO: outputs missing
            return ox.apply_invoke_inline(self.oxml.graph, arg_map, output_map)
        else:
            assert Graph.inference_runtime is not None, "No inference runtime has been provided."
            # evaluate with real values
            kwargs = {name: OnnxOperatorBuilderX.value_to_ndarray(val) for name, val in arg_map.items()}
            return Graph.inference_runtime(self.oxml, kwargs)

    def save(self, path):
        onnx.save_model(self.oxml, path)

    @staticmethod
    def load(path_or_model, name=None, inputs=None, outputs=None):
        """
        Construct a Graph object by loading an ONNX model.
        """
        oxml = onnx.load_model(path_or_model) if isinstance(path_or_model, str) else path_or_model
        for opset_import in oxml.opset_import:
            if opset_import.domain == '':
                if Graph.opset < opset_import.version:
                    raise RuntimeError("Graph opset < model opset: Graph opset = " + str(Graph.opset)
                                       + ", model opset = " + str(opset_import.version))
                elif Graph.opset > opset_import.version:
                    Graph._enforce_opset_version(oxml)
        g = Graph(name or oxml.graph.name)
        g._bind(oxml, inputs=inputs, outputs=outputs)
        return g


class Tensor(object):
    def __init__(self, tensor_name: str, ox):
        self.name = tensor_name
        self.ox = ox

    def _to_binary_tensor_args(self, other):
        # convert self, other to [self, other], but if either is a number, convert that to a constant
        x, y = self, other
        if isinstance(y, (int, float, bool, np.ndarray)):
            y = self.ox.constant(value=y)
        elif isinstance(x, (int, float, bool, np.ndarray)):
            x = self.ox.constant(value=x)
        return [x, y]

    def __add__(self, other):
        return self.ox.add(self._to_binary_tensor_args(other))

    def __sub__(self, other):
        return self.ox.sub(self._to_binary_tensor_args(other))

    def __mul__(self, other):
        return self.ox.mul(self._to_binary_tensor_args(other))

    def __div__(self, other):
        return self.ox.div(self._to_binary_tensor_args(other))

    def __pow__(self, other):
        return self.ox.pow(self._to_binary_tensor_args(other))

    def __matmul__(self, other):
        return self.ox.matmul(self._to_binary_tensor_args(other))

    def __lt__(self, other):
        return self.ox.less(self._to_binary_tensor_args(other))

    def __le__(self, other):
        return self.ox.less_or_equal(self._to_binary_tensor_args(other))

    def __eq__(self, other):
        return self.ox.equal(self._to_binary_tensor_args(other))

    def __ne__(self, other):  # ONNX has no NotEqual
        return self.ox.not_op([self.ox.equal(self._to_binary_tensor_args(other))])

    def __gt__(self, other):
        return self.ox.greater(self._to_binary_tensor_args(other))

    def __ge__(self, other):
        return self.ox.greater_or_equal(self._to_binary_tensor_args(other))

    def __neg__(self):
        return self.ox.neg([self])

    def __not__(self):
        return self.ox.not_op([self])

    def __getitem__(self, indices):
        # normalize indices to tuples of slices
        # Formats encountered:
        #  - a single int
        #  - a tuple of (int or slice)
        if not isinstance(indices, (tuple, list)):  # single item: make it a tuple
            indices = (indices,)
        squeeze = [axis for axis, index in enumerate(indices) if
                   isinstance(index, int)]  # which axes had a single index?
        indices = tuple(
            index if isinstance(index, slice) else slice(index, index + 1 if index != -1 else None, 1) for index in
            indices)  # make all tuple items of type Slice
        bs, es, ss, ds = [], [], [], []
        INT_MAX = 2 ** 63 - 1
        for axis, index in enumerate(indices):
            if not isinstance(index, slice):
                raise ValueError("Index expected")
            if index.start is None and index.stop is None:  # [:] can be skipped
                continue
            b, e, s = index.start, index.stop, index.step
            bs.append(b if b is not None else 0)
            es.append(e if e is not None else INT_MAX)
            ss.append(s if s is not None else 1)
            ds.append(axis)
        res = self.ox.slice(self, starts=bs, ends=es, axes=ds, steps=ss)
        if squeeze:  # single index means we must drop the axis
            res = self.ox.squeeze([res], axes=squeeze)
        return res

    _all_ops = [op_name[6:] for op_name in dir(onnx_ops) if op_name.startswith("apply_")] + \
               ['shape', 'constant_of_shape', 'range', 'slice', 'equal']  # these are temporarily declared here

    def __getattribute__(self, attr):
        """
        A little hack that allows to call unary operators in a chaining fashion,
        e.g. x.shape() instead of ox.shape(x).
        """
        if attr in Tensor._all_ops:
            f = self.ox.__getattribute__(attr)

            def call_it(*args, **kwargs):
                assert len(args) == 0, "In chaining expressions, only keyword args are allowed"
                assert "inputs" not in kwargs, "Chaining expressions do not currently support additional inputs"
                return f(self, *args, **kwargs)

            return call_it
        else:
            return object.__getattribute__(self, attr)


class OnnxOperatorBuilderX(OnnxOperatorBuilder):
    def _output_names_to_tensors(self, outputs):
        if isinstance(outputs, str):
            return Tensor(outputs, self)
        else:
            return tuple(self._output_names_to_tensors(output) for output in outputs)

    def _tensors_to_input_names(self, inputs):
        if isinstance(inputs, Tensor):
            return inputs.name
        else:
            return [self._tensors_to_input_names(input) for input in inputs]

    def apply_op(self, apply_func_or_op_type, inputs, name=None, outputs=None, **attrs):  # override!
        inputs = self._tensors_to_input_names(inputs)
        if isinstance(apply_func_or_op_type, str):
            return self._output_names_to_tensors(
                super().add_node(apply_func_or_op_type, inputs, name=name, outputs=outputs, **attrs))
        else:
            return self._output_names_to_tensors(
                super().apply_op(apply_func_or_op_type, inputs, name=name, outputs=outputs, **attrs))

    def apply_invoke_inline(self, ox_graph, input_map, output_map):
        input_map = dict(input_map)
        output_map = dict(output_map)
        # input_map:  [name in graph] -> actual input Tensor
        # output_map: [name in graph] -> desired name for the result, or None
        f_name = "invoke_inline_" + ox_graph.name
        for graph_output in output_map.keys():  # @TODO: use proper comprehensions
            output_map[graph_output] = self._process_outputs(
                output_map[graph_output], name=f_name)[0]
        outputs = list(output_map.values())  # remember these; these are the outputs of this invocation
        if len(outputs) == 1:
            outputs = outputs[0]  # single output
        for graph_input in input_map.keys():
            input_map[graph_input] = self._process_inputs(
                [input_map[graph_input].name], name=f_name)[0]
        _logger.debug(f_name, input_map, output_map)

        existing_node_names = {item.name: item for item in self._container.nodes}
        existing_initializer_names = {item.name: item for item in self._container.initializers}
        existing_value_infos = {item.name: item for item in self._container.value_info}

        # collect all outputs from the graph we are expanding, so that we can map them to unique names
        # @TODO: This will also map some code that may be shared later on. Leave that to the optimizer.
        node_map = dict()
        for node in ox_graph.node:
            if not node.input:  # leaves do not need to be mapped; they can just get uniq'ed
                continue
            for output in node.output:
                if output in output_map:  # this is an actual output that already has been mapped
                    continue
                uniq_name = onnx_ops._create_name_or_use_existing_one(self._scope, self._generate_name(output, None),
                                                                      None)
                output_map[output] = uniq_name
            uniq_node_name = onnx_ops._create_name_or_use_existing_one(self._scope,
                                                                       self._generate_name(node.name, None), None)
            node_map[output] = uniq_node_name

        def map_tensors(args, arg_map):
            for i in range(len(args)):
                if args[i] in arg_map:
                    _logger.debug("Remapping", args[i], "to", arg_map[args[i]])
                    args[i] = arg_map[args[i]]

        for node in ox_graph.node:
            node = copy.deepcopy(node)  # since we patch, we must clone it first
            map_tensors(node.input, input_map)  # patch the input references to the function arguments
            map_tensors(node.output, output_map)  # rename the outputs to unique ones
            map_tensors(node.input, output_map)  # outputs may be inputs to other nodes in this graph
            if node.name in node_map:
                node.name = node_map[node.name]
            if node.name in existing_node_names:
                str_node = str(node)
                str_other = str(existing_node_names[node.name])
                if str_node != str_other:
                    # must be the same, otherwise we have inconsistent dups, e.g. in input models
                    _logger.info("Duplicate node name with inconsistent nodes:\n", node, "vs:\n",
                                 existing_node_names[node.name])
                    assert str_node == str_other
                continue
            self._container.nodes.append(node)
        for initializer in ox_graph.initializer:
            if initializer.name in existing_initializer_names:  # @TODO: check if they are the same
                _logger.info("Duplicate initializer name skipped:", initializer.name)
                continue
            if initializer.name in output_map:  # technically, the whole function could be a lonely initializer
                # _logger.debug("Replacing:", initializer.name, initializer.shape)
                initializer = copy.deepcopy(initializer)
                initializer.name = output_map[initializer.name]
            # _logger.debug(initializer.name)
            self._container.initializers.append(initializer)
        for value_info in ox_graph.value_info:
            if value_info.name in existing_value_infos:  # @TODO: check if they are the same
                _logger.info("Duplicate value_info name skipped:", value_info.name)
                continue
            # @TODO: Not sure what must be mapped, and how
            _logger.debug(value_info)
            self._container.value_info.append(value_info)
        return self._output_names_to_tensors(outputs)  # note: outputs is either a string or a list of strings

    def arg(self, name):
        """
        Use this to create a function argument
        """
        return Tensor(name, self)

    # @TODO?: Should we follow the conventions in the spec? loop(self, count, cond, inputs, body)
    def loop(self, count, cond, body, inputs, outputs=None, name=None):
        inputs = self._tensors_to_input_names(inputs)
        count = None if count is None else self._tensors_to_input_names(count)
        if cond is not None:
            cond = self._tensors_to_input_names(cond)
        else:
            cond = self._tensors_to_input_names(self.constant(value=np.array([True]), name="cf"))
        sub_graph = body.oxml.graph
        return self._output_names_to_tensors(super().loop(count, cond, sub_graph,
                                                          inputs=inputs,
                                                          outputs=outputs,  # @TODO: unique output names
                                                          name=name))
